#!/usr/bin/perl -w

##############################################################################
# The purpose of this script is to compute overall statistics for the results
# summarized by mbc-evalCandGenesByCoExpr-leaveOneOut-extractDisGenes.pl
#
# copyright 2011 by Rosario M. Piro (rosario.piro@unito.it)
#
# ---------------------------------------------------------------------------
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; _exclusively_ version 2 of the License
#    NO later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################

use strict;

use Getopt::Std;
use File::Basename;
use POSIX qw(strtod);

# command buffering of STDOUT
$| = 1;

my $prgname = "mbc-evalCandGenesByCoExpr-leaveOneOut-computeStatistics.pl";

#shift(@ARGV); # all other arguments are probably summary result files

if (scalar(@ARGV)==0) {
    print STDERR "Error: at least one file must be specified (-h for help)!\n";
    exit(1);
}

if ($ARGV[0] eq "-h" || $ARGV[0] eq "--help") {
    &printUsageInfo();   
}


#       $N$ & Candidates & $g$--$p$ pairs & \multicolumn{3}{l}{Ranked first} & & \multicolumn{3}{l}{Ranked 1st--10th} & & \multicolumn{3}{l}{Ranked $\le$ 10\%} \\
#       \cline{4-6} \cline{8-10} \cline{12-14}
#        & (average) & & Obs. & Exp. & $P$-value & & Obs. & Exp. & $P$-value & & Obs. & Exp. & $P$-value \\

# HEADER
print "file\tcandidates_avg\tg-p_pairs\tranked_first_obs\tranked_first_exp\tranked_first_pval\tranked_best3_obs\tranked_best3_exp\tranked_best3_pval\tranked_best10_obs\tranked_best10_exp\tranked_best10_pval\tranked_best10perc_obs\tranked_best10perc_exp\tranked_best10perc_pval\n";
 

foreach my $resultsFile (@ARGV) {

    if (! -e $resultsFile || ! -r $resultsFile) {
	print STDERR "Error: $resultsFile does not exist or cannot be read!\n";
	exit(2);
    }

    # 1) find average number of candidates and number of cases (for cases with at leat 50 candidates and NOT unscored)

    my $avNumCands = `cat $resultsFile | grep unscored -v | filterLinesNumeric.pl -c 5 -T GE -V 50 | extractSpecColumns.pl -c 5 | buildDistributionBins.pl -b 10 | grep "average value: " | sed -e "s/^.*average value: //" | sed -e "s/[^0-9] .*\$//"`;
    chomp($avNumCands);

    my $numCases = `cat $resultsFile | grep unscored -v | filterLinesNumeric.pl -c 5 -T GE -V 50 | extractSpecColumns.pl -c 5 | buildDistributionBins.pl -b 10 | grep "total number of values: " | sed -e "s/^.*total number of values: //"`;
    chomp($numCases);

    # 2) find number of genes ranked first:

    my $n_tot;
    my $a_tot;

    my $rankedFirst = `cat $resultsFile | grep unscored -v | filterLinesNumeric.pl -c 5 -T GE -V 50 | extractSpecColumns.pl -c 6 | buildDistributionBins.pl -b 1 | grep "^1[[:space:]]" | extractSpecColumns.pl -c 2`;
    chomp($rankedFirst);

    my $rankedFirst_exp = int($numCases / $avNumCands + 0.5);

    my $rankedFirst_pval = "-"; # default: no p-value

    if ($rankedFirst eq "") { 
	$rankedFirst = "0";
    }
    else {
	$n_tot = int( $numCases*$avNumCands + 0.5 );
	$rankedFirst_pval = `probHyperGeometric -at_least -a $rankedFirst -a_tot $numCases -n $numCases -n_tot $n_tot`;
	chomp($rankedFirst_pval);
    }

    # 3.1) find number of genes ranked in top three:

    my $rankedBest3 = `cat $resultsFile | grep unscored -v | filterLinesNumeric.pl -c 5 -T GE -V 50 | extractSpecColumns.pl -c 6 | buildDistributionBins.pl -b 3 | grep "^3[[:space:]]" | extractSpecColumns.pl -c 2`;
    chomp($rankedBest3);

    my $rankedBest3_exp = int( $numCases/($avNumCands/3) + 0.5);

    my $rankedBest3_pval = "-"; # default: no p-value

    if ($rankedBest3 eq "") {
	$rankedBest3 = "0";
    }
    else {

	$n_tot = int( $numCases*$avNumCands + 0.5 );
	$a_tot = $numCases * 3;
	$rankedBest3_pval = `probHyperGeometric -at_least -a $rankedBest3 -a_tot $a_tot -n $numCases -n_tot $n_tot`;
	chomp($rankedBest3_pval);
    }

    # 3.2) find number of genes ranked in top ten:

    my $rankedBest10 = `cat $resultsFile | grep unscored -v | filterLinesNumeric.pl -c 5 -T GE -V 50 | extractSpecColumns.pl -c 6 | buildDistributionBins.pl -b 10 | grep "^10[[:space:]]" | extractSpecColumns.pl -c 2`;
    chomp($rankedBest10);

    my $rankedBest10_exp = int( $numCases/($avNumCands/10) + 0.5);

    my $rankedBest10_pval = "-"; # default: no p-value

    if ($rankedBest10 eq "") {
	$rankedBest10 = "0";
    }
    else {

	$n_tot = int( $numCases*$avNumCands + 0.5 );
	$a_tot = $numCases * 10;
	$rankedBest10_pval = `probHyperGeometric -at_least -a $rankedBest10 -a_tot $a_tot -n $numCases -n_tot $n_tot`;
	chomp($rankedBest10_pval);
    }

    # 4) find number of genes ranked in upper 10%:

    my $rankedBest10Perc = `cat $resultsFile | grep unscored -v | filterLinesNumeric.pl -c 5 -T GE -V 50 | perl -e 'while(<STDIN>) { \@tok=split(/\t/,\$_); print "".(\$tok[5]/\$tok[4])."\n";}' | buildDistributionBins.pl -b 0.1 | grep "^0.1[[:space:]]" | extractSpecColumns.pl -c 2`;
    chomp($rankedBest10Perc);

    my $rankedBest10Perc_exp = int( $numCases/10 + 0.5);

    my $rankedBest10Perc_pval = "-";

    if ($rankedBest10Perc eq "") {
	$rankedBest10Perc = "0";
    }
    else {
	$n_tot = int( $numCases*$avNumCands + 0.5 );
	$a_tot = int( $n_tot /10 + 0.5 );
	$rankedBest10Perc_pval = `probHyperGeometric -at_least -a $rankedBest10Perc -a_tot $a_tot -n $numCases -n_tot $n_tot`;
	chomp($rankedBest10Perc_pval);
    }


#       $N$ & Candidates & $g$--$p$ pairs & \multicolumn{3}{l}{Ranked first} & & \multicolumn{3}{l}{Ranked 1st--10th} & & \multicolumn{3}{l}{Ranked $\le$ 10\%} \\
#       \cline{4-6} \cline{8-10} \cline{12-14}
#        & (average) & & Obs. & Exp. & $P$-value & & Obs. & Exp. & $P$-value & & Obs. & Exp. & $P$-value \\

    if ($rankedFirst_pval ne "-") {
	$rankedFirst_pval = sprintf("%.2e",$rankedFirst_pval);
    }
    if ($rankedBest3_pval ne "-") {
	$rankedBest3_pval = sprintf("%.2e",$rankedBest3_pval);
    }
    if ($rankedBest10_pval ne "-") {
	$rankedBest10_pval = sprintf("%.2e",$rankedBest10_pval);
    }
    if ($rankedBest10Perc_pval ne "-") {
	$rankedBest10Perc_pval = sprintf("%.2e",$rankedBest10Perc_pval);
    }

    my $file = basename($resultsFile);
    print "$file\t".sprintf("%.1f",$avNumCands)."\t$numCases\t$rankedFirst\t$rankedFirst_exp\t$rankedFirst_pval\t$rankedBest3\t$rankedBest3_exp\t$rankedBest3_pval\t$rankedBest10\t$rankedBest10_exp\t$rankedBest10_pval\t$rankedBest10Perc\t$rankedBest10Perc_exp\t$rankedBest10Perc_pval\n";


}

exit(0);

#############################################################################
#                                   FUNCTIONS:
#############################################################################


# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# --------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario M. Piro (rosario.piro\@unito.it)\n\n";
    print "Usage:\n";
    print "$prgname [-h] FILES\n";
    print "\nWhere FILES are the summary result files returned by the script\n";
    print "mbc-evalCandGenesByCoExpr-leaveOneOut-extractDisGenes.pl\n\n";
    print "-h                Display this help message and exit.\n\n";

    exit(0);

} # printUsageInfo() ...
