/*****************************************************************************
 * evalCandGenesByCoExpr-serviceroutines.cpp - CGPACE - evaluation of        *
 * candidate genes, given a list of candidates, a list of known disease and  *
 * a list of Pearson correlation coefficients (or a similar measure for      *
 * co-expression; having higher values for positive co-expression).          *
 * Version 0.0.6                                                             *
 * ------------------------------------------------------------------------- *
 * Copyright 2007-2012 by Rosario Piro (mailto: rosario.piro@unito.it;       *
 *                                                     r.piro@dkfz.de)       *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; _exclusively_ version 2 of the License    *
 *   NO later version.                                                       *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

//---------------------
//--- include files ---
//---------------------
#include <fstream>
#include <iostream>
#include <map>
#include <set>
#include <vector>
//#include <iomanip>
#include <cmath>
#include <cstring>
#include <stdlib.h>
//#include <values.h>
#include <errno.h>
#include "extstdlib.h"
//#include <time.h>
#include "evalCandGenesByCoExpr-serviceroutines.h"

#include <climits>
#define NO_INDEX UINT_MAX


//------------------------------/
//--- parseDisGeneListFile() ---/
//------------------------------/

int parseDisGeneListFile(const string inFName, set<string> &disGenes4L1O) {
  int retVal = 0;

  // for parsing the input file
  char inBuffer[IBL_+1];

  #ifdef DEBUG
  cout << "#DEBUG: Parsing input file: " << inFName << endl;
  #endif

  // open the file
  ifstream inFile(inFName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open file '" << inFName << "'"
	 << " for reading." << endl;
    return 128;      // error
  } // if(!inFile) ...

  while (inFile.getline(inBuffer,IBL_,'\n')) {

    string line = inBuffer;

    #ifdef DEBUG
    //cout << "#DEBUG: parsing line: '" << inBuffer << "'" << flush << endl;
    #endif

    if (line.compare(0,1,"#",0,1)==0)
      continue; // commented line!

    // each line should have the form: "<gene_id>\t<OMIM_id>\t..."
    // trying to parse it:
    vector<string> lineTokens;
    if (splitString(line, '\t', lineTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: cannot split string!" << endl;
      #endif

      return 129;
    }
    if (lineTokens.size() < 2) {

      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: Wrong number of fields! "
	   << "Found: " << lineTokens.size() << ", expected: at least 2." << endl;
      #endif

      return 130;
    }

    // keep this in the list for the leave-one-out
    disGenes4L1O.insert(lineTokens[0]+"\t"+lineTokens[1]);

  } // while (inFile.getline(...

  
  inFile.close(); // parsing done.

  return retVal; // ok?

} // parseDisGeneListFile() ...



//---------------------------/
//--- parseGeneListFile() ---/
//---------------------------/


int parseGeneListFile(const string inFName,
		      vector<pair<string, string> > &chrGeneList,
		      map<string,unsigned int> &gene2ListPos) {

  int retVal = 0;

  unsigned int geneIDPos = NO_INDEX;
  unsigned int chrPos = NO_INDEX;
  bool firstLine = true;

  // for parsing the input file
  char inBuffer[IBL_+1];

  #ifdef DEBUG
  cout << "#DEBUG: Parsing input file: " << inFName << endl;
  #endif

  // open the file
  ifstream inFile(inFName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open file '" << inFName << "'"
	 << " for reading." << endl;
    return 128;      // error
  } // if(!inFile) ...

  while (inFile.getline(inBuffer,IBL_,'\n')) {

    string line = inBuffer;

    #ifdef DEBUG
    cout << "#DEBUG: parsing line: '" << inBuffer << "'" << flush << endl;
    #endif

    if (line.compare(0,1,"#",0,1)==0)
      continue; // commented line!

    // each line should have the form: "<gene_id>\t<similarOmim>\t<realOmim>\t..."
    // trying to parse it:
    vector<string> lineTokens;
    if (splitString(line, '\t', lineTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: cannot split string!" << endl;
      #endif

      return 129;
    }

    // try to find positions of necessary data columns
    if (firstLine) {

      unsigned int ii;
      for (ii = 0; ii < lineTokens.size(); ii++) {
	if (lineTokens[ii] == GENEID_COLUMN_NAME)
	  geneIDPos = ii;
	else if (lineTokens[ii] == CHR_COLUMN_NAME)
	  chrPos = ii;
      }

      if (geneIDPos == NO_INDEX || chrPos == NO_INDEX) {
	// we didn't find all necessary columns, complain!!!

	cerr << "Error: file " << inFName << " must contain the columns \""
	     << GENEID_COLUMN_NAME << "\" and \"" << CHR_COLUMN_NAME << "\"!"
	     << endl << flush;
	return 144;
      }

      #ifdef DEBUG
      cout << "#DEBUG: got header; gene ID at index " << geneIDPos
	   << " (column " << (geneIDPos+1) << "); chromosome at index "
	   << chrPos << " (column " << (chrPos+1) << ")." << endl << flush;
      #endif

      firstLine = false;
      continue; // don't process the header further, we got what we need
    }


    if (geneIDPos+1 > lineTokens.size() || chrPos+1 > lineTokens.size()) {

      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: Wrong number of fields! "
	   << "Found: " << lineTokens.size() << ", required columns are "
	   << (geneIDPos+1) << " and " << (chrPos+1) << "." << endl;
      #endif

      return 130;
    }

    string thisGene = lineTokens[geneIDPos];
    string thisChr = lineTokens[chrPos];

    // add this chromosome/gene-pair to the ordered list:
    pair<string,string> newChrGPair;
    newChrGPair.first = thisChr;
    newChrGPair.second = thisGene;

    chrGeneList.push_back(newChrGPair);

    // keep track of it's position in the list:
    pair<string, unsigned int> newGPosPair;
    newGPosPair.first = thisGene;
    newGPosPair.second = chrGeneList.size()-1;

    if (gene2ListPos.insert(newGPosPair).second != true) {
      cerr << "Error: couldn't insert gene " <<  thisGene << " on chromosome "
	   << thisChr << " from file " << inFName << " to gene list." << endl
	   << flush;
      exit(157);
    }

    #ifdef DEBUG
    cout << "#DEBUG: added gene " << newGPosPair.first << " on chromosome "
	 << newChrGPair.first << " at vector index " << newGPosPair.second
	 << endl << flush;
    #endif

  } // while ( ...

  inFile.close();

  return retVal; // ok?

} // parseGeneListFile() ...



//-------------------------/
//--- parseWeightFile() ---/
//-------------------------/
int parseWeightFile(const string inFName, vector<double> &weights){
  int retVal = 0;

  // for parsing the input file
  char inBuffer[IBL_+1];

  #ifdef DEBUG
  cout << "#DEBUG: Parsing input file: " << inFName << endl;
  #endif

  // open the file
  ifstream inFile(inFName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open file '" << inFName << "'"
	 << " for reading." << endl;
    return 128;      // error
  } // if(!inFile) ...




  #ifdef DEBUG
  int lineCount = 0;
  #endif

  while (inFile.getline(inBuffer,IBL_,'\n')) {

    string line = inBuffer;

    #ifdef DEBUG
    //cout << "#DEBUG: parsing line: '" << inBuffer << "'" << flush << endl;
    lineCount++;
    #endif

    if (line.compare(0,1,"#",0,1)==0)
      continue; // commented line!

    // each line can contain one or more tab-separated weights.
    // Only numeric values are considered as weights, the rest will be ignored
    // trying to parse it:
    vector<string> lineTokens;
    if (splitString(line, '\t', lineTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: cannot split string!" << endl;
      #endif

      return 229;
    }

    unsigned int v;
    for (v = 0; v < lineTokens.size(); v++) {
      // for each weight:
      if (isNumeric(lineTokens[v].c_str())) {
	weights.push_back(atof(lineTokens[v].c_str()));
      }
      else {
        #ifdef DEBUG
	cout << "#DEBUG: Warning: field " << (v+1) << " in input line "
	     << lineCount
	     << " is not numeric ... ignoring!" << endl << flush;
        #endif
      }
    }

  } // while (inFile.getline(...
  
  inFile.close(); // parsing done.

  return retVal; // ok?

} // parseWeightFile() ...




//-------------------------------/
//--- parseGene2OmimMapFile() ---/
//-------------------------------/

int parseGene2OmimMapFile(const string inFName,
			  map<string,set<string> > &omim2SimPhenoMap,
			  map<string,set<string> > &omim2GenesMap) {
  int retVal = 0;

  // for parsing the input file
  char inBuffer[IBL_+1];

  #ifdef DEBUG
  cout << "#DEBUG: Parsing input file: " << inFName << endl;
  #endif

  // open the file
  ifstream inFile(inFName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open file '" << inFName << "'"
	 << " for reading." << endl;
    return 128;      // error
  } // if(!inFile) ...

  while (inFile.getline(inBuffer,IBL_,'\n')) {

    string line = inBuffer;

    #ifdef DEBUG
    cout << "#DEBUG: parsing line: '" << inBuffer << "'" << flush << endl;
    #endif

    if (line.compare(0,1,"#",0,1)==0)
      continue; // commented line!

    // each line should have the form: "<gene_id>\t<similarOmim>\t<realOmim>\t..."
    // trying to parse it:
    vector<string> lineTokens;
    if (splitString(line, '\t', lineTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: cannot split string!" << endl;
      #endif

      return 129;
    }
    if (lineTokens.size() < 3) {

      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: Wrong number of fields! "
	   << "Found: " << lineTokens.size() << ", expected: at least 3." << endl;
      #endif

      return 130;
    }

    string relDisGene = lineTokens[0];
    string queryOmim = lineTokens[1];
    string relOmim = lineTokens[2];

    // first add the related disease to omim2GenesMap, if we haven't it already)
    if (omim2GenesMap.find(relOmim) == omim2GenesMap.end()) {

      // add a new item to the map:
      pair<string,set<string> > newPair;
      newPair.first = relOmim;

      if ((omim2GenesMap.insert(newPair)).second != true) {
	retVal = 124;
        #ifdef DEBUG
	cout << "#DEBUG: Error: couldn't add map item of "
	     << newPair.first << " to omim2GenesMap!" << endl;
        #endif
      }
      else {
        #ifdef DEBUG
	cout << "#DEBUG: Mapping omim2GenesMap of: " << newPair.first
	     << endl;
        #endif
      }
    }

    // now add the disease gene to omim2GenesMap
    if ((*(omim2GenesMap.find(relOmim))).second.insert(relDisGene).second) {
      #ifdef DEBUG
      cout << "#DEBUG: OMIM " << relOmim << " -> gene " << relDisGene
	   << endl;
      #endif
    }

    // now do the same for the similar phenotype
    if (omim2SimPhenoMap.find(queryOmim) == omim2SimPhenoMap.end()) {

      // add a new item to the map:
      pair<string,set<string> > newPair;
      newPair.first = queryOmim;

      if ((omim2SimPhenoMap.insert(newPair)).second != true) {
	retVal = 124;
        #ifdef DEBUG
	cout << "#DEBUG: Error: couldn't add map item of "
	     << newPair.first << " to omim2SimPhenoMap!" << endl;
        #endif
      }
      else {
        #ifdef DEBUG
	cout << "#DEBUG: Mapping omim2SimPhenoMap of: " << newPair.first
	     << endl;
        #endif
      }
    }

    if ((*(omim2SimPhenoMap.find(queryOmim))).second.insert(relOmim).second) {
      #ifdef DEBUG
      cout << "#DEBUG: OMIM " << queryOmim << " -> similar OMIM " << relOmim
	   << endl;
      #endif
    }

  } // while (inFile.getline(...

  
  inFile.close(); // parsing done.

  return retVal; // ok?

} // parseGene2OmimMapFile() ...





//-----------------------/
//--- parseGeneFile() ---/
//-----------------------/

int parseGeneFile(const string inFName, vector<string> &genes) {

  int retVal = 0;

  // for parsing the input file
  char inBuffer[IBL_+1];

  set<string> readGeneIds;

  #ifdef DEBUG
  cout << "#DEBUG: Parsing input file: " << inFName << endl;
  #endif

  // open the file
  ifstream inFile(inFName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open file '" << inFName << "'"
	 << " for reading." << endl;
    return 28;      // error
  } // if(!inFile) ...

  while (inFile.getline(inBuffer,IBL_,'\n')) {

    string line = inBuffer;

    #ifdef DEBUG
    //cout << "#DEBUG: parsing line: '" << inBuffer << "'" << flush << endl;
    #endif

    if (line.compare(0,1,"#",0,1)==0)
      continue; // commented line!

    // each line should have the form: "<gene_id>" or "<gene_id>\t<whatever>"
    // trying to parse it:
    vector<string> lineTokens;
    if (splitString(line, '\t', lineTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: cannot split string!" << endl;
      #endif

      return 29;
    }
    if (lineTokens.size() < 1) {

      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: Wrong number of fields! "
	   << "Found: " << lineTokens.size() << ", expected: at least 1." << endl;
      #endif

      return 30;
    }

    if (readGeneIds.insert(lineTokens[0]).second) {
      // this gene was not yet present, insert it.
      genes.push_back(lineTokens[0]);

      #ifdef DEBUG
      cout << "#DEBUG: Gene: " << lineTokens[0] << endl;
      #endif

    }
    #ifdef DEBUG
    else {
      cout << "#DEBUG: Warning: already having gene "
	   << lineTokens[0] << "; ignoring!" << endl;
    }
    #endif

  } // while (inFile.getline(...

  inFile.close(); // parsing done.

  return retVal; // ok?

} // parseGeneFile() ...



// -------------------------------
// --- addScore2CoexprList() ---
// -------------------------------

int addScore2CoexprList(map<string,map<string,double> > &coexprListsScoreMap,
			  const string refGene, const string pGene,
			  const double coexScore) {

  int retVal = 0;

  map<string,map<string,double> >::iterator rGPMit =
    coexprListsScoreMap.find(refGene);

  // refGene is a reference gene! Add this CoexScore to it's map?
  if (rGPMit != coexprListsScoreMap.end()) {

    map<string,double>::iterator pMit = ((*rGPMit).second).find(pGene);
    if (pMit == ((*rGPMit).second).end()) {
      // not yet present, add!
      pair<string,double> newPair;
      newPair.first = pGene;
      newPair.second = coexScore;
      bool inserted = (((*rGPMit).second).insert(newPair)).second;
      if (!inserted) {
	retVal = 22;
        #ifdef DEBUG
	cout << "#DEBUG: Error: couldn't add coexpression score " << coexScore
	     << " of gene " << pGene << " for co-expression list base gene " 
	     << refGene << "!" << endl << flush;
	#endif
      }
      else {
	#ifdef DEBUG
	cout << "#DEBUG: Adding co-expression score " << newPair.second
	     << " of gene " << newPair.first << " for co-expression list base "
	     << "gene " << refGene << "." << endl << flush;
	#endif
      }

    } // if (pMit == ...
    else { 
      #ifdef DEBUG
      cout << "#DEBUG: Not adding co-expression score of gene " << pGene
	   << " for co-expression list base gene " << refGene
	   << "; already present!" << endl << flush;
      #endif
    }

  } // if (rGPMit != ...
  else {
    #ifdef DEBUG
    cout << "#DEBUG: Not adding co-expression score of gene " << pGene
	 << " for co-expression list base gene " << refGene 
	 << "; no such base gene!" << endl << flush;
    #endif
  }

  return retVal;
} // addScore2CoexprList()


//-------------------------------------/
//--- buildCoexprLists4BaseGenes () ---/
//-------------------------------------/

int buildCoexprLists4BaseGenes(map<string,vector<double> > &exprProfileMap,
			    double (*computeCoExprTwoGenes)(vector<double> &,
							    vector<double> &,
							    vector<double> &,
							    void *),
			 map<string,map<string,double> > &coexprListsScoreMap,
			 vector<double> &weights,
			 void *functSpecInput) {

  int retVal = 0;

  map<string,map<string,double> >::iterator coexprListIt = 
    coexprListsScoreMap.begin();

  while (coexprListIt != coexprListsScoreMap.end()) {
    // for each base gene (either candidates or reference genes):

    string baseGene = (*coexprListIt).first;

    map<string,vector<double> >::iterator baseGeneExprIt =
      exprProfileMap.find(baseGene);

    if (baseGeneExprIt != exprProfileMap.end()) {
      // compute co-expression scores for all genes with expression profiles:
      map<string,vector<double> >::iterator gExprIt;
      for (gExprIt = exprProfileMap.begin(); gExprIt != exprProfileMap.end();
	   gExprIt++) {

	string thisGene = (*gExprIt).first;

	double coexScore = computeCoExprTwoGenes((*baseGeneExprIt).second,
						 (*gExprIt).second,
						 weights, functSpecInput);

        #ifdef DEBUG
	//cout << "#DEBUG: Co-expression score for " << baseGene 
	//   << " (base gene) and " << thisGene << ": " << coexScore << endl;
        #endif
      
	// add coexScore to coexprList:
	retVal = addScore2CoexprList(coexprListsScoreMap, baseGene, 
				     thisGene, coexScore);

      } // for (gExprIt = ...
    }
    else {
      #ifdef DEBUG
      cout << "#DEBUG: No expression profile for base gene " << baseGene << endl;
      #endif
    }


    coexprListIt++;

  } // while (coexprListsScoreMap != ...

  return retVal; // ok?

} // buildCoexprLists4BaseGenes() ...



// ---------------------------------
// --- parseExpressionDataFile() ---
// ---------------------------------

int parseExpressionDataFile(const string inFName,
			    set<string> &genesWithExprData,
			    map<string,vector<double> > &exprProfileMap,
			    vector<string> &headerInfo) {

  int retVal = 0;

  // for parsing the input file
  char inBuffer[IBL_+1];

  unsigned int lineCount = 0;

  #ifdef DEBUG
  cout << "#DEBUG: Parsing input file: " << inFName << endl << flush;
  #endif
  
  // open the file
  ifstream inFile(inFName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open file '" << inFName << "'"
	 << " for reading." << endl;
    return 41;      // error
  } // if(!inFile) ...

  while (inFile.getline(inBuffer,IBL_,'\n')) {

    string line = inBuffer;

    #ifdef DEBUG
    //cout << "#DEBUG: parsing line: '" << inBuffer << "'" << flush << endl;
    #endif

    lineCount++;

    if (line.compare(0,1,"#",0,1)==0)
      continue; // commented line!

    if (line.compare(0,8,"probeset",0,8)==0
	|| line.compare(0,1,"\t",0,1)==0
	|| line.compare(0,9,"region_id",0,9)==0
	|| line.compare(0,5,"COORD",0,5)==0 
	|| line.compare(0,4,"gene\t",0,5)==0 
	) {
      // header line containing tissue info, keep it!

      if (splitString(line, '\t', headerInfo) != 0) {
        #ifdef DEBUG
	cout << "#DEBUG: Error while parsing header line: cannot split string!" << endl;
        #endif

	return 47;
      }

      continue; // next line, this is not an expression profile!
    }

    // each line should have the form: "<gene_id>\t<exprVal1>\t<exprVal2>..."
    // trying to parse it:
    vector<string> lineTokens;
    if (splitString(line, '\t', lineTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: cannot split string!" << endl;
      #endif

      return 42;
    }
    if (lineTokens.size() < 2) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: Wrong number of fields! "
	   << "Found: " << lineTokens.size() << ", expected: at least 2 "
	   << endl;
      #endif

      return 43;
    }

    // now try to convert:
    string gene = lineTokens[0];

    // get expression values:
    vector<double> expressionVals;

    unsigned int v;
    for (v = 1; v < lineTokens.size(); v++) {
      // for each expression value:
      if (isNumeric(lineTokens[v].c_str())) {
	expressionVals.push_back(atof(lineTokens[v].c_str()));
      }
      else {
        #ifdef DEBUG
	cout << "#DEBUG: Error: field " << (v+1) << " in input line " << lineCount
	     << " is not numeric!" << endl << flush;
        #endif

	retVal = 23;
      }
    }

    // now add expression profile to map:
    pair<string,vector<double> > newPair;
    newPair.first = gene;
    newPair.second = expressionVals;

    if ((exprProfileMap.insert(newPair)).second != true) {
      retVal = 24;
      #ifdef DEBUG
      cout << "#DEBUG: Error: couldn't add expression profile of gene "
	   << newPair.first << " to map!" << endl;
      #endif
    }
    else {
      #ifdef DEBUG
      cout << "#DEBUG: Mapping expression profile of gene: " << newPair.first
	   << endl;
      #endif
    }

    // keep track of the fact that this gene has expression data:
    if (genesWithExprData.find(gene) == genesWithExprData.end()) {
      if ((genesWithExprData.insert(gene)).second != true) {
	cerr << "Error: couldn't flag gene " << gene
	     << " for having expression data!" << endl;
	exit(44);
      }
    }


  } // while (inFile.getline(...



  inFile.close(); // parsing done.

  return retVal; // ok?

} // parseExpressionDataFile() ...




//--------------------------/
//--- parseDGPDataFile() ---/
//--------------------------/

int parseDGPDataFile(const string inFName, map<string,double> &geneDGPmap) {

  int retVal = 0;

  int lineNumber = 0;

  // for parsing the input file
  char inBuffer[IBL_+1];

  #ifdef DEBUG
  cout << "#DEBUG: Parsing input file: " << inFName << endl;
  #endif

  // open the file
  ifstream inFile(inFName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open file '" << inFName << "'"
	 << " for reading." << endl;
    return 44;      // error
  } // if(!inFile) ...

  while (inFile.getline(inBuffer,IBL_,'\n')) {

    string line = inBuffer;

    lineNumber++;

    #ifdef DEBUG
    //cout << "#DEBUG: parsing line: '" << inBuffer << "'" << flush << endl;
    #endif

    if (line.compare(0,1,"#",0,1)==0)
      continue; // commented line!

    // each line should have the form: "<gene>\t<DGPprob>..."
    // trying to parse it:
    vector<string> lineTokens;
    if (splitString(line, '\t', lineTokens) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing line: cannot split string!" << endl;
      #endif

      return 45;
    }

    if (lineTokens.size() < 2) {
      // not enough fields in this line, complain and ignore!
      #ifdef DEBUG
      cout << "#DEBUG: Warning: Wrong number of fields! "
	   << "Found: " << lineTokens.size() << ", expected: at least 2"
	   << endl;
      #endif

      continue;
    }

    // now try to convert:
    if (isNumeric(lineTokens[1].c_str())) {

      pair<string,double> newPair;
      newPair.first = lineTokens[0];                // gene ID
      newPair.second = atof(lineTokens[1].c_str()); // probability

      if ((geneDGPmap.insert(newPair)).second != true) {
	retVal = 25;
	#ifdef DEBUG
	cout << "#DEBUG: Error: couldn't add " << newPair.first << "->"
	     << newPair.second << " to gene -> DGP prob. map!" << endl;
	#endif
      }
      else {
	#ifdef DEBUG
	cout << "#DEBUG: Mapping " << newPair.first << " -> "
	     << newPair.second << endl;
	#endif
      }

    }
    else {
      #ifdef DEBUG
      cout << "#DEBUG: Error: line " << lineNumber << " requires numeric value in second field (DGP probability)!" << endl;
      #endif
    }

  } // while (inFile.getline(...

  inFile.close(); // parsing done.

  return retVal; // ok?

} // parseDGPDataFile() ...



// -------------------------
// --- sortCoexprLists() ---
// -------------------------

int sortCoexprLists(map<string,map<string,double> > &coexprListsScoreMap,
		    map<string,map<string,double> > &coexprListsRankMap) {
  int retVal = 0;

  map<string,map<string,double> >::iterator coexprListsIt =
    coexprListsScoreMap.begin();

  while (coexprListsIt != coexprListsScoreMap.end()) {
    // iteration over all co-expression lists!

    string baseGene = (*coexprListsIt).first;
    
    int vecLen = ((*coexprListsIt).second).size();

    #ifdef DEBUG
    cout << "#DEBUG: Ranking coexpr. list (" << vecLen
	 << " genes) for base gene " << baseGene << endl << flush;
    #endif

    // create corresponding rank list!
    map<string,double> newRankMap;

    if (vecLen != 0) {
      // sort only if list not empty!

      // sort according to CoexScore:
      map<string,double>::iterator *pArray =
	new map<string,double>::iterator[vecLen];

      // copying iterators into array:
      int i = 0;
      map<string,double>::iterator pIter = ((*coexprListsIt).second).begin();
      while (pIter != ((*coexprListsIt).second).end()) {
	pArray[i++] = pIter++;
      }

      // now sort the array:
      qsort(pArray, vecLen, sizeof(map<string,double>::iterator),
	    compareStringDoubleMapValsDesc);

      // determine resulting ranks:
      int rank;
      for (rank = 1; rank <= vecLen; rank++) {
	pair<string,double> newPair;
	newPair.first = (*(pArray[rank-1])).first;    // gene
	newPair.second = (double)rank/(double)vecLen; // relative rank

	if ( (newRankMap.insert(newPair)).second != true ) {
	  #ifdef DEBUG
	  cout << "#DEBUG: WARNING: couldn't add relative rank " << newPair.second
	       << " of gene " << newPair.first
	       << " to base gene " << baseGene
	       << endl << flush;
          #endif

	  retVal = 26;
	}
	else {
	  #ifdef DEBUG
	  map<string,map<string,double> >::iterator it1 =
	    coexprListsScoreMap.find(baseGene);
	  map<string,double>::iterator it2 =
	    ((*it1).second).find(newPair.first);
	  double coexScore = (*it2).second;
	  cout << "#DEBUG: For base gene: " << baseGene << "; gene "
	       << newPair.first << " has relative rank " << newPair.second
	       << " (for co-expr. score:  " << coexScore << ")" << endl << flush;
	  #endif
	}
      } // for (rank ...

      delete pArray;

    } // if (vecLen != ...

    // now we have a corresponding rank map, try to insert:
    pair<string,map<string,double> > newPair;
    newPair.first = baseGene;                     // base gene
    newPair.second = newRankMap;                  // rank map

    if ( (coexprListsRankMap.insert(newPair)).second != true ) {
      #ifdef DEBUG
      cout << "#DEBUG: ERROR: couldn't add (relative) rank map to base gene "
	   << baseGene << endl << flush;
      #endif

      retVal = 27;
    }

    coexprListsIt++;  // next co-expression list = base gene

  } // while (coexprListsIt != ...

  return retVal; // ok?
} // sortCoexprLists() ...



// ----------------------------------------
// --- compareStringDoubleMapValsDesc() ---
// ----------------------------------------
// to be used for qsort()!
int compareStringDoubleMapValsDesc(const void *pIt1, const void *pIt2) {

  map<string,double>::iterator *pIter1 = (map<string,double>::iterator *)pIt1;
  map<string,double>::iterator *pIter2 = (map<string,double>::iterator *)pIt2;

  double diff = (*(*pIter1)).second - (*(*pIter2)).second;

  if (diff<0)
    return 1;      // note: this is inverted to what requested for qsort
  else if (diff>0) // since we want the CoexScores to be ordered in descending
    return -1;   // order!

  return 0; // seems to be the same ...

} // compareStringDoubleMapValsDesc()...


// ----------------------------------------
// --- compareUnsignedInt2DoubleDesc() ---
// ----------------------------------------
// to be used for sort()!
bool compareUnsignedInt2DoubleDesc(const pair<unsigned int, double> &v1, const pair<unsigned int, double> &v2) {

  double diff = v1.second - v2.second;

  //cout << "comparing: " << v1.second << " (pos=" << v1.first << ") with " << v2.second << " (pos=" << v2.first << "); diff=" << diff << endl;

  if (diff>0)
    return true;      // note: this is inverted to what requested for sort
  else if (diff<0)    // since we want the expression levles to be ordered in 
    return false;     // descending order!

  return false; // seems to be the same ...

} // compareUnsignedInt2DoubleDesc()...


// ----------------------------------------------------------
// --- compareString2DoubleDesc(), compareString2Double() ---
// ----------------------------------------------------------
// to be used for sort()!
bool compareString2DoubleDesc(const pair<string, double> &v1, const pair<string, double> &v2) {

  double diff = v1.second - v2.second;

  //cout << "comparing: " << v1.second << " (pos=" << v1.first << ") with " << v2.second << " (pos=" << v2.first << "); diff=" << diff << endl;

  if (diff>0)
    return true;
  else if (diff<0)
    return false;

  return false; // seems to be the same ...

} // compareString2DoubleDesc()...


bool compareString2Double(const pair<string, double> &v1, const pair<string, double> &v2) {

  double diff = v2.second - v1.second;

  //cout << "comparing: " << v2.second << " (pos=" << v2.first << ") with " << v1.second << " (pos=" << v1.first << "); diff=" << diff << endl;

  if (diff>0)
    return true;
  else if (diff<0)
    return false;

  return false; // seems to be the same ...

} // compareString2Double()...



// ----------------------------
// --- rankCandidateGenes() ---
// ----------------------------

int rankCandidateGenes(map<string,double> &candGene2ScoreMap,
		       vector<string> &rankedCandGenes) {

  int retVal = 0;

  int vecLen = candGene2ScoreMap.size();

  #ifdef DEBUG
  cout << "#DEBUG: Ranking candidate genes (" << vecLen
       << " genes)." << endl << flush;
  #endif

  if (vecLen != 0) {
    // sort only if list not empty!

    // sort according to CoexScore:
    map<string,double>::iterator *cArray =
      new map<string,double>::iterator[vecLen];

    // copying iterators into array:
    int i = 0;
    map<string,double>::iterator cIter = candGene2ScoreMap.begin();
    while (cIter != candGene2ScoreMap.end()) {
      cArray[i++] = cIter++;
    }

    // now sort the array:
    qsort(cArray, vecLen, sizeof(map<string,double>::iterator),
	  compareStringDoubleMapVals);

    // determine resulting ranks:
    int rank;
    for (rank = 1; rank <= vecLen; rank++) {

      rankedCandGenes.push_back((*cArray[rank-1]).first);

      #ifdef DEBUG
      cout << "#DEBUG: Candidate gene " << (*cArray[rank-1]).first
	   << " with score ";
      if ((*cArray[rank-1]).second != UNSCORED)
	cout  << (*cArray[rank-1]).second;
      else
	cout << "UNSCORED";
      cout << " has rank: " << rank << endl << flush;
      #endif
    } // for (rank ...
    
    delete cArray;

  } // if (vecLen != ...

  return retVal; // ok?

} // rankCandidateGenes() ...



// ------------------------------------------------
// --- evaluateCandidateList4ReferenceGeneSet() ---
// ------------------------------------------------
int evaluateCandidateList4ReferenceGeneSet(vector<string> &candGenes, vector<string> &refGenes,
					   map<string,vector<double> > &exprProfileMap, 
					   vector<double> &weights, void *functSpecInfo,
					   bool useDGPprob, map<string,double> &geneDGPmap,
					   bool coexprLists4CandGenes,
					   map<string,map<string,double> > &coexprListsRankMap,
					   double (*computeCoExprTwoGenes)(vector<double> &,
									   vector<double> &,
									   vector<double> &,
									   void *),
					   bool isRandomization, bool extendedOutput,
					   // output:
					   vector<string> &rankedCandGenes,
					   map<string,double> &candGene2ScoreMap,
					   map<string,double> &candGene2AvCoexScoreMap,
					   // extended output:
					   map<string,vector<double> > &candGene2relPositionsMap,
					   // alternative output for randomization:
					   vector<double> &randScores) {

  int retVal = 0;

  map<string,map<string,double> > coexprListsScoreMap;
                               // maps coexpr. lists to a gene-CoexScore map

  // compute co-expression scores and build coexpr. lists for base genes
  #ifdef DEBUG
  cout << "#DEBUG: Evaluating " << candGenes.size() << " candidate genes for " << refGenes.size() << " reference genes";
  if (coexprLists4CandGenes) {
    cout << " (building co-expression lists for candidate genes!)";
  }
  cout << endl;
  #endif


  // do we use co-expression lists for candidates instead of ref. genes?
  vector<string> *exprListBaseGenes = &refGenes;
  if (coexprLists4CandGenes) {
    exprListBaseGenes = &candGenes;
  }

  // for each reference gene (or candidate gene, if exprListBaseGenes==true), initialize an unordered coexpr. list:
  unsigned int g;
  for (g = 0; g < (*exprListBaseGenes).size(); g++) {

    if (coexprListsRankMap.find((*exprListBaseGenes)[g]) != coexprListsRankMap.end()) {
      // we already have the relative ranks for this gene; don't compute again!
      continue;
    }

    // we didn't have the gene before! compute the relative rank lists:
    map<string,double> newGene2CoexScoreMap;
    pair<string,map<string,double> > newPair;
    newPair.first = (*exprListBaseGenes)[g];    // base gene ID (cand or ref)
    newPair.second = newGene2CoexScoreMap;          // map for genes and coexScores

    bool inserted = (coexprListsScoreMap.insert(newPair)).second;

    if (!inserted) {
      cerr << "Error: could not create map for co-expression scores of co-expression list base gene " << (*exprListBaseGenes)[g] << " ... quitting!" << endl << flush;
      exit(111);
    }
  }


  if (coexprListsScoreMap.size() > 0) { // we need relative ranks for at least some of the base genes
    // compute co-expression scores and build coexpr. lists for base genes
    #ifdef DEBUG
    cout << "#DEBUG: === Need to build coexpression lists for some of the genes ..." << endl;
    #endif

    if ( (retVal=buildCoexprLists4BaseGenes(exprProfileMap,
					    computeCoExprTwoGenes,
					    coexprListsScoreMap,
					    weights, functSpecInfo)) != 0) {
      cerr << "Error: Couldn't build coexpression lists ... quitting! "
	   << "Errorcode: " << retVal << endl;
      exit(retVal);
    }



    // now start sorting the co-expression lists:
    #ifdef DEBUG
    cout << "#DEBUG: === Sorting co-expression lists ... " << endl;
    #endif

    if ( (retVal=sortCoexprLists(coexprListsScoreMap,
				 coexprListsRankMap)) != 0) {
      cerr << "Error: Couldn't correctly sort the co-expression lists"
	   << " ... quitting! Errorcode: " << retVal << endl;
      exit(retVal);
    }

  } // if (coexprListsScoreMap.size > 0 ...


  // we have the ranks of all genes in the co-expression lists, now we can
  // compute scores for the candidate genes!
  #ifdef DEBUG
  cout << "#DEBUG: === Computing overall scores for candidates ..." << endl;
  #endif


  for (g = 0; g < candGenes.size(); g++) {
    // iteration over all candidate genes

    #ifdef DEBUG
    cout << "#DEBUG: Computing score for candidate " << candGenes[g] << endl;
    #endif

    // compute score and average relative rank:
    double thisScore = UNSCORED;

    double sumRelRank = 0.0;
    int numRelRanks = 0;

    // either overall rank prob of this candidate on all reference coexpression
    // lists (default); or rank prob of all reference genes agains this
    // candidate's co-expression list (option -C on command line).
    vector<string> searchGenes;
    vector<string> coexprListGenes;
    if (!coexprLists4CandGenes) { // default
      searchGenes.push_back(candGenes[g]);
      coexprListGenes = refGenes;
    }
    else { // option -C
      searchGenes = refGenes;
      coexprListGenes.push_back(candGenes[g]);
    }

    // now get scores:
    vector<string>::iterator sGit;
    vector<string>::iterator cLGit;
    for (sGit = searchGenes.begin(); sGit != searchGenes.end(); sGit++) {
      // for each search gene (either this candidate or all reference genes):

      for (cLGit = coexprListGenes.begin(); cLGit != coexprListGenes.end();
	   cLGit++) {
	// for each co-expression list (either this gene's or those of all
	// reference genes):

	// find rank if we have one (maybe not all genes have expression data):
	map<string,map<string,double> >::iterator ceLRMit =
	  coexprListsRankMap.find(*cLGit);

	if (ceLRMit != coexprListsRankMap.end()) {
	  // we have a co-expression list (might also be empty).
	  map<string,double>::iterator rankMapIt = 
	    ((*ceLRMit).second).find(*sGit);

	  if (rankMapIt != ((*ceLRMit).second).end()) {
	    // we have a relative rank for this search gene in the coexpr.list!
	    double thisRank = (*rankMapIt).second;

	    if (thisScore == UNSCORED)
	      thisScore = 1; // multipying this with all relative ranks gives
	                     // an overall probability of having such ranks
	                     // by chance!
	    thisScore *= thisRank;

	    sumRelRank += thisRank;
	    numRelRanks++;

	    // keep track of single relative positions, if extended output:
	    if (extendedOutput) {
	      string cG;
	      if (!coexprLists4CandGenes) { // default
		cG = *sGit;
	      }
	      else {
		cG = *cLGit;
	      }

	      map<string,vector<double> >::iterator rPit = 
		candGene2relPositionsMap.find(cG);

	      // is this candidate gene already in the map?
	      if (rPit == candGene2relPositionsMap.end()) { // not yet in map

		vector<double> newVec;
		pair<string,vector<double> > newPair;
		newPair.first = cG;
		newPair.second = newVec;

		if ((candGene2relPositionsMap.insert(newPair)).second !=true) {
		  cerr << "Error: couldn't add " << newPair.first
		       << " to map of relative positions in co-expression "
		       << "lists!" << endl;
                  exit(55);
		}

		rPit = candGene2relPositionsMap.find(cG);
	      }

	      // add the relative position to the vector!
	      (*rPit).second.push_back(thisRank);

	      //#ifdef DEBUG
	      //cout << "#DEBUG: added relative position (" << thisRank
	      //   << ") for search gene " << (*sGit) << " in co-expression "
	      //   << "list of gene " << (*cLGit) << " to relative position "
	      //   << "map of candidate gene " << cG << endl;
	      //#endif

	    } // if (extendedOutput) ...


	    #ifdef DEBUG
	    cout << "#DEBUG: Search gene " << (*sGit) << " has relative rank "
		 << thisRank
		 << " in co-expression list of base gene " << (*cLGit)
		 << " [score=" << thisScore << "]" << endl;
            #endif
	  }
	  else {
	    #ifdef DEBUG
	    cout << "#DEBUG: No relative rank for search gene " << (*sGit)
		 << " in co-expression list of base gene " << (*cLGit)
		 << "!" << endl;
            #endif
	  }
	}
	else {
	  #ifdef DEBUG
	  cout << "#DEBUG: No co-expression list for base gene " << (*cLGit)
	       << "!" << endl;
          #endif
	}

      } // for (cLGit = ...

    } // for (sGit = ...

    #ifdef DEBUG
    cout << "#DEBUG: Overall score for candidate gene " << candGenes[g]
	 << ": ";
    if (thisScore == UNSCORED)
      cout << "UNSCORED";
    else
      cout << thisScore;
    cout << endl << flush;
    #endif


    if (useDGPprob) {
      map<string,double>::iterator gDGPit = geneDGPmap.find(candGenes[g]);
      if (gDGPit == geneDGPmap.end()) {
	#ifdef DEBUG
	cout << "#DEBUG: No DGP prob. for gene " << candGenes[g]
	     << "! Considering as unscored!" << endl;
        #endif

	thisScore = UNSCORED;
      }
      else { // got a DGP prob.
	double dgp = (*gDGPit).second;
	double overallScore =  thisScore * (1.0 - dgp);
	                          // multiply with score from co-expr. lists,
	                          // but keep in mind that higher DGP prob.
	                          // = more likely to be a disease gene (while
	                          // we have lower scores for better candidates

	#ifdef DEBUG
	cout << "#DEBUG: DGP prob. for gene " << candGenes[g]
	     << ": " << dgp << " => overall score: " << thisScore 
	     << " * (1.0 - " << dgp << ") = " << overallScore << endl;
        #endif

	thisScore = overallScore;
      }

    }


    if (!isRandomization) { // need not only scores but mappings gene->score
      // insert into score map!
      pair<string,double> newPair;
      newPair.first = candGenes[g];
      newPair.second = thisScore;

      if ( (candGene2ScoreMap.insert(newPair)).second != true) {
	cerr << "Error: couldn't insert overall score of candidate "
	     << candGenes[g] << " into score map! Quitting!" << endl;
	exit(56);
      }

      // compute average relative rank!
      double avRelRank = UNSCORED;
      if (numRelRanks != 0) {
	avRelRank = sumRelRank / (double) numRelRanks;

        #ifdef DEBUG
	cout << "#DEBUG: Average relative rank of candidate gene " << candGenes[g]
	     << ": " // << sumRelRank << " / " << numRelRanks << " = "
	     << avRelRank << endl << flush;
        #endif

	// insert into map:
	pair<string,double> newPair;
	newPair.first = candGenes[g];
	newPair.second = avRelRank;

	if ( (candGene2AvCoexScoreMap.insert(newPair)).second != true) {
	  cerr << "Error: couldn't insert average relative rank of candidate " 
	       << candGenes[g] << " into score map! Quitting!" << endl;
	  exit(57);
	}
      }
    }
    else { // this is for randomization; only keep the score, we don't need more info ...
      randScores.push_back(thisScore);
    }

  } // for (g = ...


  if (!isRandomization) {
    // now rank candidates (for randomizations this is done _after_ all iterations, not here):
    #ifdef DEBUG
    cout << "#DEBUG: === Ranking candidates ..." << endl;
    #endif

    if ( (retVal=rankCandidateGenes(candGene2ScoreMap,
				    rankedCandGenes)) != 0) {
      cerr << "Error: Couldn't correctly rank candidate genes"
	   << " ... quitting! Errorcode: " << retVal << endl;
      exit(retVal);
    }
  }



  return retVal; // ok?


  } // evaluateCandidateList4ReferenceGeneSet() ...



// ------------------------------------
// --- compareStringDoubleMapVals() ---
// ------------------------------------
// to be used for qsort()!
int compareStringDoubleMapVals(const void *pIt1, const void *pIt2) {

  map<string,double>::iterator *pIter1 = (map<string,double>::iterator *)pIt1;
  map<string,double>::iterator *pIter2 = (map<string,double>::iterator *)pIt2;

  double diff = (*(*pIter1)).second - (*(*pIter2)).second;

  if (diff>0)
    return 1;
  else if (diff<0)
    return -1;

  return 0; // seems to be the same ...

} // compareStringDoubleMapVals()...



/* removed this; using the weighted version instead, see below!
// ------------------------------------------
// --- scoring functions based on Pearson ---
// ------------------------------------------

double __computePearson(vector<double> &exprProfile1,
			vector<double> &exprProfile2,
			bool normalize) {

  double pearson = NOPEARSON;

  unsigned int vecLen = exprProfile1.size();
  unsigned int vecLen2 = exprProfile2.size();

  if (vecLen != vecLen2) {
    #ifdef DEBUG
    cout << "#DEBUG: Warning: expression profiles have different length ("
	 << vecLen << " and " << vecLen2 << "); using minimum length." << endl;
    #endif

    if (vecLen2 < vecLen)
      vecLen = vecLen2; // use only vecLen from now on
  }

  if (vecLen > 0) {
    double sum_E1 = 0.0;
    double sum_E2 = 0.0;
    double sum_E1xE1 = 0.0;
    double sum_E2xE2 = 0.0;
    double sum_E1xE2 = 0.0;

    for (unsigned int eV = 0; eV < vecLen; eV++) {
      double exprVal1 = exprProfile1[eV];
      double exprVal2 = exprProfile2[eV];

      sum_E1 += exprVal1;
      sum_E2 += exprVal2;
      sum_E1xE2 += exprVal1 * exprVal2;
      sum_E1xE1 += exprVal1 * exprVal1;
      sum_E2xE2 += exprVal2 * exprVal2;
    }

    pearson = ( sum_E1xE2 - sum_E1*sum_E2/(double)vecLen )
              / ( sqrt( (sum_E1xE1 - sum_E1*sum_E1/(double)vecLen)
		  *(sum_E2xE2 - sum_E2*sum_E2/(double)vecLen) ) );
  }


  if (normalize) {
    pearson *= sqrtl(vecLen - 1);

    #ifdef DEBUG
    //cout << "#DEBUG: Normalized Pearson correlation coefficient: " << pearson
    // << endl << flush;
    #endif
  }
  else {
    #ifdef DEBUG
    //cout << "#DEBUG: Pearson correlation coefficient: " << pearson 
    // << endl << flush;
    #endif
  }

  return pearson;

} // __computePearson() ...


double computePearson(vector<double> &exprProfile1,
		      vector<double> &exprProfile2) {

  return __computePearson(exprProfile1, exprProfile2, false);

} // computePearson() ...

double computeQuadraticPearson(vector<double> &exprProfile1,
			       vector<double> &exprProfile2) {

  double r = __computePearson(exprProfile1, exprProfile2, false);

  return r*r;

} // computePearson() ...

double computeNormalizedPearson(vector<double> &exprProfile1,
				vector<double> &exprProfile2) {

  return __computePearson(exprProfile1, exprProfile2, true);

} // computeNormalizedPearson() ...

// removed because useing the weighted version, see below! 
*/


// -----------------------------------------------------------------------
// --- scoring functions based on both WEIGHTED and UNWEIGHTED Pearson ---
// -----------------------------------------------------------------------

double __computePearson(vector<double> &exprProfile1,
			vector<double> &exprProfile2,
			vector<double> &weights, bool normalize) {

  double pearson = NOPEARSON;

  unsigned int vecLen = exprProfile1.size();
  unsigned int vecLen2 = exprProfile2.size();
  unsigned int weightLen = weights.size();

  bool useWeights = true;

  if (weightLen == 0)
    useWeights = false;

  if (vecLen != vecLen2) {
    #ifdef DEBUG
    cout << "#DEBUG: Warning: expression profiles have different length ("
	 << vecLen << " and " << vecLen2 << "); using minimum length." << endl;
    #endif

    if (vecLen2 < vecLen)
      vecLen = vecLen2; // use only vecLen from now on
  }

  if (useWeights && weightLen < vecLen) {
    cerr << "Error: weight vector has length " << weightLen
	 << "; expression vector requires length " << vecLen << "!" 
	 << endl << flush;
    return NOPEARSON;
  }

  if (vecLen > 0) {
    double sum_weight = 0.0;
    double sum_wE1 = 0.0;
    double sum_wE2 = 0.0;
    double sum_wE1xE1 = 0.0;
    double sum_wE2xE2 = 0.0;
    double sum_wE1xE2 = 0.0;

    for (unsigned int eV = 0; eV < vecLen; eV++) {
      double exprVal1 = exprProfile1[eV];
      double exprVal2 = exprProfile2[eV];

      double weight = 1.0;
      if (useWeights)
	weight = weights[eV];

      sum_weight += weight;
      sum_wE1 += weight * exprVal1;
      sum_wE2 += weight * exprVal2;
      sum_wE1xE2 += weight * exprVal1 * exprVal2;
      sum_wE1xE1 += weight * exprVal1 * exprVal1;
      sum_wE2xE2 += weight * exprVal2 * exprVal2;
    }

    if(sum_wE1 == 0.0 || sum_wE2 == 0.0)
      pearson = 0.0; // if one of the vectors has no expression at all
    else
      pearson = ( sum_wE1xE2 - sum_wE1*sum_wE2/sum_weight )
                / ( sqrt( (sum_wE1xE1 - sum_wE1*sum_wE1/sum_weight)
	  	    *(sum_wE2xE2 - sum_wE2*sum_wE2/sum_weight) ) );
  }


  if (normalize) {
    pearson *= sqrtl(vecLen - 1);

    //#ifdef DEBUG
    //cout << "#DEBUG: Normalized Pearson correlation coefficient: " << pearson
    // << endl << flush;
    //#endif
  }
  else {
    //#ifdef DEBUG
    //cout << "#DEBUG: Pearson correlation coefficient: " << pearson 
    // << endl << flush;
    //#endif
  }

  return pearson;

} // __computeWeightedPearson() ...


double computePearson(vector<double> &exprProfile1,
		      vector<double> &exprProfile2) {

  vector<double> weights; // empty dummy vector

  return __computePearson(exprProfile1, exprProfile2, weights, false);

} // computePearson() ...

double computePearson(vector<double> &exprProfile1,
		      vector<double> &exprProfile2,
		      vector<double> &weights,
		      void *functSpecInput) {

  return __computePearson(exprProfile1, exprProfile2, weights, false);

} // computePearson() ...


double computeQuadraticPearson(vector<double> &exprProfile1,
			       vector<double> &exprProfile2) {

  vector<double> weights; // empty dummy vector

  double r = __computePearson(exprProfile1, exprProfile2, weights, false);

  return r*r;

} // computePearson() ...

double computeQuadraticPearson(vector<double> &exprProfile1,
			       vector<double> &exprProfile2,
			       vector<double> &weights,
			       void *functSpecInput) {

  double r = __computePearson(exprProfile1, exprProfile2, weights, false);

  return r*r;

} // computePearson() ...


double computeNormalizedPearson(vector<double> &exprProfile1,
				vector<double> &exprProfile2) {

  vector<double> weights; // empty dummy vector

  return __computePearson(exprProfile1, exprProfile2, weights, true);

} // computeNormalizedPearson() ...

double computeNormalizedPearson(vector<double> &exprProfile1,
				vector<double> &exprProfile2,
				vector<double> &weights,
				void *functSpecInput) {

  return __computePearson(exprProfile1, exprProfile2, weights, true);

} // computeNormalizedPearson() ...



// -----------------------------------------
// --- computeRelativeIntensityOverlap() ---
// -----------------------------------------

double computeRelativeIntensityOverlap(vector<double> &exprProfile1,
				       vector<double> &exprProfile2,
				       vector<double> &weights,
				       void *functSpecInput) {

  // (double *)funcSpecInput is the minimum normalized expression value (0..1);
  // lower expression values will be considered as noise (i.e. ignored).

  double minNormExpr = DBL_MIN; // default: no minimum!

  if (functSpecInput != NULL) {
    // minimum normalized expression value is specified!
    minNormExpr = *((double *)functSpecInput);
  }


  double relIntOv = NORELINTOV;

  unsigned int vecLen = exprProfile1.size();
  unsigned int vecLen2 = exprProfile2.size();
  unsigned int weightLen = weights.size();

  bool useWeights = true;

  if (weightLen == 0)
    useWeights = false;

  if (vecLen != vecLen2) {
    #ifdef DEBUG
    cout << "#DEBUG: Warning: expression profiles have different length ("
	 << vecLen << " and " << vecLen2 << "); using minimum length." << endl;
    #endif

    if (vecLen2 < vecLen)
      vecLen = vecLen2; // use only vecLen from now on
  }

  if (useWeights && weightLen < vecLen) {
    cerr << "Error: weight vector has length " << weightLen
	 << "; expression vector requires length " << vecLen << "!" 
	 << endl << flush;
    return NORELINTOV;
  }



  if (vecLen > 0) {
    double sumIntOv = 0.0;
    double sumMaxIntOv = 0.0;

    /* CHANGE/START;
    // DON'T NORMALIZE; SHOULD BE DONE BEFORE, DIRECTLY ON EXPRESSION DATA!

    // first: get maximum and minimum expression levels of both vectors,
    // so that we can normalize all to values between 0 and 1
    double maxExp1 = DBL_MIN;
    double minExp1 = DBL_MAX;
    double maxExp2 = DBL_MIN;
    double minExp2 = DBL_MAX;

    for (unsigned int eV = 0; eV < vecLen; eV++) {
      if (exprProfile1[eV] > maxExp1)
	maxExp1 = exprProfile1[eV];
      if (exprProfile1[eV] < minExp1)
	minExp1 = exprProfile1[eV];

      if (exprProfile2[eV] > maxExp2)
	maxExp2 = exprProfile2[eV];
      if (exprProfile2[eV] < minExp2)
	minExp2 = exprProfile2[eV];
    }

    //#ifdef DEBUG
    //cout << "#DEBUG: in computeRelativeIntensityOverlap: determined min/max expression levels for the two profiles: minExp1=" << minExp1 << "; maxExp1=" << maxExp1 << "; minExp2=" << minExp2 << "; maxExp2=" << maxExp2 << endl;
    //#endif

    // if we still have a positive minimum, assume zero
    if (minExp1 > 0) {
      minExp1 = 0;

      //#ifdef DEBUG
      //cout << "#DEBUG: in computeRelativeIntensityOverlap: setting minExp1=0" << endl;
      //#endif
    }

    if (minExp2 > 0) {
      minExp2 = 0;

      //#ifdef DEBUG
      //cout << "#DEBUG: in computeRelativeIntensityOverlap: setting minExp2=0" << endl;
      //#endif
    }


    for (unsigned int eV = 0; eV < vecLen; eV++) {

      double weight = 1.0;
      if (useWeights)
	weight = weights[eV];

      sum_weight =+ weight;

      double shiftedRelExpr1 = (exprProfile1[eV] + -1*minExp1) /
	                        (maxExp1 + -1*minExp1);
      double shiftedRelExpr2 = (exprProfile2[eV] + -1*minExp2) /
	                        (maxExp2 + -1*minExp2);

    NOT NORMALIZING ANYMORE ... DIRECTLY TAKING EXPRESSION VALUES: */
    for (unsigned int eV = 0; eV < vecLen; eV++) {
      double weight = 1.0;
      if (useWeights)
	weight = weights[eV];

      double shiftedRelExpr1 = exprProfile1[eV];
      double shiftedRelExpr2 = exprProfile2[eV];
    /* CHANGE/END */


      // if a minimum normalized expression value is specified:
      if (shiftedRelExpr1 < minNormExpr)
	shiftedRelExpr1 = 0.0;
      if (shiftedRelExpr2 < minNormExpr)
	shiftedRelExpr2 = 0.0;


      double thisIntOv = shiftedRelExpr1 * shiftedRelExpr2;

      double thisMaxIntOv = shiftedRelExpr1 * shiftedRelExpr1; // always >0
      if (abs(shiftedRelExpr2) > abs(shiftedRelExpr1))
	thisMaxIntOv = shiftedRelExpr2 * shiftedRelExpr2; // always >0
      
      sumIntOv += weight * thisIntOv;
      sumMaxIntOv += weight * thisMaxIntOv;

      //#ifdef DEBUG
      //cout << "#DEBUG: in computeRelativeIntensityOverlap: intensity overlap for index eV=" << eV << " of expression profiles (with weight=" << weight << "):" << endl
      //   << "#DEBUG: exprProfile1[eV]=" << exprProfile1[eV] << " => shiftedRelExpr1=" << shiftedRelExpr1 << "; exprProfile2[eV]=" << exprProfile2[eV] << " => shiftedRelExpr2=" << shiftedRelExpr2 << endl
      //   << "#DEBUG: => thisIntOv=" << thisIntOv << "; thisMaxIntOv=" << thisMaxIntOv << "; sums now: sumIntOv=" << sumIntOv << "; sumMaxIntOv=" << sumMaxIntOv << endl << flush;
      //#endif

    }

    if (sumMaxIntOv == 0 && sumIntOv == 0) // no expression at all
      relIntOv = 1.0; // both images/profiles are empty, but they are the same!
    else if (sumMaxIntOv != 0) {
      relIntOv = sumIntOv / sumMaxIntOv;
    }

    //#ifdef DEBUG
    //cout << "#DEBUG: in computeRelativeIntensityOverlap: relative intensity overlap = " <<  relIntOv
    // << endl << flush;
    //#endif
  }

  return relIntOv;

} // computeRelativeIntensityOverlap() ...

double computeRelativeIntensityOverlap(vector<double> &exprProfile1,
			     vector<double> &exprProfile2) {

  vector<double> weights; // empty dummy vector

  return computeRelativeIntensityOverlap(exprProfile1, exprProfile2, weights,
					 NULL);
} // computeRelativeIntensityOverlap() ...






// --------------------------------------------------
// --- computeEucDist() , computeInverseEucDist() ---
// --------------------------------------------------

double computeEucDist(vector<double> &exprProfile1,
			     vector<double> &exprProfile2,
			     vector<double> &weights,
			     void *functSpecInput) {

  double eucDist = NOEUCDIST;

  unsigned int vecLen = exprProfile1.size();
  unsigned int vecLen2 = exprProfile2.size();
  unsigned int weightLen = weights.size();

  bool useWeights = true;

  if (weightLen == 0)
    useWeights = false;

  if (vecLen != vecLen2) {
    #ifdef DEBUG
    cout << "#DEBUG: Warning: expression profiles have different length ("
	 << vecLen << " and " << vecLen2 << "); using minimum length." << endl;
    #endif

    if (vecLen2 < vecLen)
      vecLen = vecLen2; // use only vecLen from now on
  }

  if (useWeights && weightLen < vecLen) {
    cerr << "Error: weight vector has length " << weightLen
	 << "; expression vector requires length " << vecLen << "!" 
	 << endl << flush;
    return NOEUCDIST;
  }



  if (vecLen > 0) {
    double sum = 0.0;
    double sum_weight = 0.0;

    for (unsigned int eV = 0; eV < vecLen; eV++) {

      double weight = 1.0;
      if (useWeights)
	weight = weights[eV];

      sum_weight =+ weight;

      double thisFact = exprProfile1[eV] - exprProfile2[eV];
      sum += weight * thisFact * thisFact;
    }

    sum = sum/sum_weight;
    eucDist = sqrt(sum);      

    #ifdef DEBUG
    //cout << "#DEBUG: Euclidean distance: " <<  invEucDist
    // << endl << flush;
    #endif
  }

  return eucDist;

} // computeEucDist() ...

double computeEucDist(vector<double> &exprProfile1,
		      vector<double> &exprProfile2) {

  vector<double> weights; // empty dummy vector

  return computeEucDist(exprProfile1, exprProfile2, weights, NULL);
} // computeEucDist() ...


double computeInverseEucDist(vector<double> &exprProfile1,
			     vector<double> &exprProfile2,
			     vector<double> &weights,
			     void *functSpecInput) {
  double invEucDist;
  double eucDist = computeEucDist(exprProfile1, exprProfile2, weights, functSpecInput);
  if (eucDist == 0)
    invEucDist = INVEUCDISTZERO; // zero distance, use maximum possible score for inverse
  else
    invEucDist = 1.0 / eucDist;
  return invEucDist;
} // computeInverseEucDist() ...


double computeInverseEucDist(vector<double> &exprProfile1,
			     vector<double> &exprProfile2) {

  double invEucDist;
  double eucDist = computeEucDist(exprProfile1, exprProfile2);
  if (eucDist == 0)
    invEucDist = INVEUCDISTZERO; // zero distance, use maximum possible score for inverse
  else
    invEucDist = 1.0 / eucDist;
  return invEucDist;
} // computeInverseEucDist() ...



// -----------------------
// --- computeCosine() ---
// -----------------------

double computeCosine(vector<double> &exprProfile1,
		     vector<double> &exprProfile2,
		     vector<double> &weights,
		     void *functSpecInput) {

  double cosine = NOCOSINE;

  unsigned int vecLen = exprProfile1.size();
  unsigned int vecLen2 = exprProfile2.size();
  unsigned int weightLen = weights.size();

  bool useWeights = true;

  if (weightLen == 0)
    useWeights = false;

  if (vecLen != vecLen2) {
    #ifdef DEBUG
    cout << "#DEBUG: Warning: expression profiles have different length ("
	 << vecLen << " and " << vecLen2 << "); using minimum length." << endl;
    #endif

    if (vecLen2 < vecLen)
      vecLen = vecLen2; // use only vecLen from now on
  }

  if (useWeights && weightLen < vecLen) {
    cerr << "Error: weight vector has length " << weightLen
	 << "; expression vector requires length " << vecLen << "!" 
	 << endl << flush;
    return NOCOSINE;
  }



  if (vecLen > 0) {
    double sumAxB = 0.0;
    double sumAxA = 0.0;
    double sumBxB = 0.0;

    for (unsigned int eV = 0; eV < vecLen; eV++) {

      double weight = 1.0;
      if (useWeights)
	weight = weights[eV];

      sumAxB += weight * exprProfile1[eV] * exprProfile2[eV];

      sumAxA += weight * exprProfile1[eV] * exprProfile1[eV];
      sumBxB += weight * exprProfile2[eV] * exprProfile2[eV];
    }

    if (sumAxA*sumBxB != 0) {
      cosine = sumAxB / (sqrt(sumAxA) * sqrt(sumBxB));
    }
    else
      cosine = NOCOSINE; // at least one of the vectors is zero, there is no angle ...

    #ifdef DEBUG
    //cout << "#DEBUG: (Weighted) cosine similarity = cosine between expr. vectors: " <<  cosine
    // << endl << flush;
    #endif
  }

  return cosine;

} // computeCosine() ...

double computeCosine(vector<double> &exprProfile1,
		     vector<double> &exprProfile2) {

  vector<double> weights; // empty dummy vector

  return computeCosine(exprProfile1, exprProfile2, weights, NULL);
} // computeCosine() ...




// --------------------------
// --- computePearson3D() ---
// --------------------------

double computePearson3D(vector<double> &exprProfile1,
			vector<double> &exprProfile2,
			vector<double> &weights, // weights for planes!
			void *functSpecInput
			) {

  // functSpecInput points to the list of columns associated to the single
  // planes of the 3D (or multi-D) space; build this object with
  // convertHeaderInfo2Planes

  map<string,vector<unsigned int> > *planes
    = (map<string,vector<unsigned int> > *)functSpecInput;

  double sumPearson = 0;

  if (planes == NULL || (*planes).size() == 0) {
    return NOPEARSON3D;
  }


  #ifdef DEBUG
  cout << "#DEBUG: computing 3D Pearson ..." << endl << flush;
  #endif


  unsigned int plCount = 0;
  map<string,vector<unsigned int> >::iterator plIt;
  for (plIt = (*planes).begin(); plIt != (*planes).end(); plIt++) {

    plCount++;

    #ifdef DEBUG
    cout << "#DEBUG: plane: " << (*plIt).first << flush;
    #endif


    // build temporary exprProfiles for this plane:
    vector<double> exprPlane1;
    vector<double> exprPlane2;

    vector<unsigned int>::iterator plposIt;
    for (plposIt = ((*plIt).second).begin();
	 plposIt != ((*plIt).second).end();
	 plposIt++) {

      #ifdef DEBUG
      cout << "; col " << *plposIt << " [" << exprProfile1[*plposIt]
           << " vs. " << exprProfile2[*plposIt] << "]";
      #endif

      exprPlane1.push_back(exprProfile1[*plposIt]);
      exprPlane2.push_back(exprProfile2[*plposIt]);

    }

    // get Pearson for this plane
    double thisPearson = computePearson(exprPlane1, exprPlane2);


    #ifdef DEBUG
    cout << " => pearson: " << thisPearson
	 << flush;
    #endif


    // weight it?
    if (weights.size() != 0) {
      thisPearson *= weights[plCount];
      #ifdef DEBUG
      cout << "; weight: " << weights[plCount];
      #endif
    }

    // now add this to the sum:
    sumPearson += thisPearson;

    #ifdef DEBUG
    cout << " => current sum: " << sumPearson << endl << flush;
    #endif
  }


  double avPearson = (sumPearson/(*planes).size());

  #ifdef DEBUG
  cout << "#DEBUG: number of planes: " << (*planes).size()
       << " => average (= 3D) Pearson: " << avPearson << endl << flush;
  #endif

  return avPearson; // average Pearson

} // computePearson3D() ...



// ------------------------------
// - convertHeaderInfo2Planes() -
// ------------------------------

int convertHeaderInfo2Planes(vector<string> &headerInfo,
			     map<string,vector<unsigned int> > &planes) {

  unsigned int hidx;
  for (hidx = 1; hidx < headerInfo.size(); hidx++) {

    #ifdef DEBUG
    cout << "#DEBUG: header of expression column " << (hidx) << ": " << headerInfo[hidx] << endl;
    #endif


    vector<string> coords;
    if (splitString(headerInfo[hidx], ',', coords) != 0) {
      #ifdef DEBUG
      cout << "#DEBUG: Error while parsing header information: cannot split string!" << endl;
      #endif

      return 48;
    }

    unsigned int cidx;
    for (cidx = 0; cidx < coords.size(); cidx++) {
      string plane = "";
      plane.append(int2string(cidx+1));
      plane.append("=");
      plane.append(coords[cidx]);

      // if we don't have a corresponding plain yet:
      if (planes.find(plane) == planes.end()) {
	vector<unsigned int> newPlaneVec;

	pair<string,vector<unsigned int> > newPair;
	newPair.first = plane;
	newPair.second = newPlaneVec;
	bool inserted = (planes.insert(newPair)).second;
	if (!inserted) {
          #ifdef DEBUG
	  cout << "#DEBUG: Error: couldn't add plane " << plane
	       << " from header info!" << endl << flush;
          #endif

	  return 49;
	}
	else {
	  #ifdef DEBUG
	  cout << "#DEBUG: new plane in header info: " << plane << endl
	       << flush;
	  #endif
	}
      }

      // insert this header index in the list of coordinates for this plane
      (*planes.find(plane)).second.push_back(hidx-1); // -1 because first column is ID!

      #ifdef DEBUG
      cout << "#DEBUG: added header index " << (hidx-1) << " to plane "
	   << plane << "; now having: " << (*planes.find(plane)).second.size() << endl << flush;
      #endif
    }
  }

  return 0;
} // convertHeaderInfo2Planes() ...



// ---------------------------------
// - computeAveragePairwiseScore() -
// ---------------------------------

double computeAveragePairwiseScore(vector<string> &genes,
			   map<string,vector<double> > &exprProfileMap,
			   double (*computeCoExprTwoGenes)(vector<double> &,
							   vector<double> &,
							   vector<double> &,
							   void *),
			   vector<double> &weights,
			   void *functSpecInput) {

  unsigned int ii = 0;
  unsigned int kk = 0;

  int count = 0;
  double sum = 0;
  double average = 0;

  vector<double> exProfile1;
  vector<double> exProfile2;

  for (ii = 0; ii < genes.size(); ii++) {
    for (kk = ii+1; kk < genes.size(); kk++) {
      count++;

      exProfile1 = (*(exprProfileMap.find(genes[ii]))).second;
      exProfile2 = (*(exprProfileMap.find(genes[kk]))).second;

      sum += computeCoExprTwoGenes(exProfile1, exProfile2, weights,
				   functSpecInput);
    }
  }


  if (count != 0) {
    average = sum/(double)count;
  }
  else {
    average = DBL_MIN; // no average since no gene pairs
  }

  // pedantic check:
  int expCount = genes.size() * (genes.size()-1) / 2;
  if (count != expCount) {
    cerr << "Error: Unexpected number of pairwise correlation scores for "
	 << genes.size() << " genes. Expected: " << expCount << ", got: "
	 << count << endl << flush;
    exit(113);
  }

  return average;

} // computeAveragePairwiseScore() ...




int getStdOutFromCommand(const string command, string &out) {

  int retVal = 0;

  out = ""; // start new string


#ifdef NO_POPEN
  // this is to avoid strange memmory problems with popen on ramon ...
  // usually not needed!


  // for parsing the input file
  char inBuffer[IBL_+1];
  string tmpFileName = "/tmp/_getstdoutfromcommand.tmp";

  #ifdef DEBUG
  cout << "#DEBUG: Getting stdout (NO_POPEN) from command: " << command << endl;
  #endif

  string cmd = command + " > " + tmpFileName;

  retVal = system(cmd.c_str());
  if (retVal != 0) {
    cerr << "Error: cannot run external command: " << cmd << endl;
    if (retVal == -1) {
      cerr << "Error from system(): errno=" << errno << "; msg="
	   << strerror(errno) << endl;
    }
    return retVal;
  }

  // open the file
  ifstream inFile(tmpFileName.c_str());
  if(!inFile) { // couldn't open
    cerr << "Error: cannot open tmp file '" << tmpFileName << "'"
	 << " for reading." << endl;
    return 117;      // error
  } // if(!inFile) ...

  while (inFile.getline(inBuffer,IBL_,'\n')) {
    out.append(inBuffer);
    out.append("\n");
  }

  inFile.close(); // parsing done.

#else
  // this is the usual procedure ... popen should be fine

  FILE *stream;
  char buff[IBL_+1];

  #ifdef DEBUG
  cout << "#DEBUG: Getting stdout from command: " << command << endl;
  #endif

  stream = popen(command.c_str(), "r");
  if (!stream) {
    cerr << "Error: cannot run external command: " << command << endl;
    cerr << "Error from popen: errno=" << errno << "; msg="
	 << strerror(errno) << endl;
    return errno;
  }

  while (!feof(stream)){
    if (fgets(buff, IBL_+1, stream) != NULL){
      out.append(buff);
    }
  }

  pclose(stream);


#endif


  return retVal;
}



/*
template <class T>
bool addToVectorUnique(vector<T> &vect, const T toAdd) {

  bool found = false;
  vector<T>::iterator vIt = vect.begin();
  while (!found && vIt != vect.end()) {

    if (*vIt == toAdd)
      found = true;

    vIt++;
  }

  if (found)
    return false;
  else
    vect.push_back(toAdd);

  return true;

}
*/
