/*****************************************************************************
 * evalCandGenesByCoExpr-serviceroutines.h - CGPACE - evaluation of          *
 * candidate genes, given a list of candidates, a list of known disease and  *
 * a list of Pearson correlation coefficients (or a similar measure for      *
 * co-expression; having higher values for positive co-expression).          *
 * Version 0.0.6                                                             *
 * ------------------------------------------------------------------------- *
 * Copyright 2007-2012 by Rosario Piro (mailto: rosario.piro@unito.it;       *
 *                                              r.piro@dkfz.de)              *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; _exclusively_ version 2 of the License    *
 *   NO later version.                                                       *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

#ifndef _CGPACE_SERVICEROUTINES_H_
#define _CGPACE_SERVICEROUTINES_H_

//---------------------
//--- include files ---
//---------------------
#include <map>
#include <set>
#include <vector>
#include <string>
#include <values.h>

//---------------------------
//--- preprocessor macros ---
//---------------------------
#define IBL_ 4194304       // input buffer length for reading files
#define DEFAULT_PRECISION 15 // for the output of floating point numbers
#define UNSCORED DBL_MAX
#define NOPEARSON DBL_MIN
#define NOPEARSON3D DBL_MIN
#define NOEUCDIST DBL_MIN
#define NOCOSINE DBL_MIN
//#define NOSPEARMAN DBL_MIN
#define NORELINTOV DBL_MIN
#define INVEUCDISTZERO DBL_MAX

#define GENEID_COLUMN_NAME "Ensembl Gene ID"
#define CHR_COLUMN_NAME "Chromosome Name" 

using namespace std;


//-----------------------------/
//--- function declarations ---/
//-----------------------------/

int parseGeneFile(const string inFName, vector<string> &genes);

int parseWeightFile(const string inFName, vector<double> &weights);

int parseGeneListFile(const string inFName,
		      vector<pair<string, string> > &chrGeneList,
		      map<string,unsigned int> &gene2ListPos);

int parseExpressionDataFile(const string inFName,
			    set<string> &genesWithExprData,
			    map<string,vector<double> > &exprProfileMap,
			    vector<string> &headerInfo);

int convertHeaderInfo2Planes(vector<string> &headerInfo,
			     map<string,vector<unsigned int> > &planes);

int parseGene2OmimMapFile(const string inFName,
			  map<string, set<string> > &omim2SimPhenoMap,
			  map<string, set<string> > &omim2GenesMap);

int parseDisGeneListFile(const string inFName, set<string> &disGenes4L1O);

int parseDGPDataFile(const string inFName, map<string,double> &geneDGPmap);

int buildCoexprLists4BaseGenes(map<string,vector<double> > &exprProfileMap,
			     double (*computeCoExprTwoGenes)(vector<double> &,
							     vector<double> &,
							     vector<double> &,
							     void *),
			 map<string,map<string,double> > &coexprListsScoreMap,
			 vector<double> &weights,
			 void *functSpecInput);

int addScore2CoexprList(map<string,map<string,double> > &coexprListsScoreMap,
			const string refGene, const string pGene,
			const double coexScore);

int sortCoexprLists(map<string,map<string,double> > &coexprListsScoreMap,
		    map<string,map<string,double> > &coexprListsRankMap);

int rankCandidateGenes(map<string,double> &candGene2ScoreMap,
		       vector<string> &rankedCandGenes);


int evaluateCandidateList4ReferenceGeneSet(vector<string> &candGenes, vector<string> &refGenes,
					   map<string,vector<double> > &exprProfileMap, 
					   vector<double> &weights, void *functSpecInfo,
					   bool useDGPprob, map<string,double> &geneDGPmap,
					   bool coexprLists4CandGenes,
					   map<string,map<string,double> > &coexprListsRankMap,
					   double (*computeCoExprTwoGenes)(vector<double> &,
									   vector<double> &,
									   vector<double> &,
									   void *),
					   bool isRandomization, bool extendedOutput,
					   // output:
					   vector<string> &rankedCandGenes,
					   map<string,double> &candGene2ScoreMap,
					   map<string,double> &candGene2AvCoexScoreMap,
					   // extended output:
					   map<string,vector<double> > &candGene2relPositionsMap,
					   // alternative output for randomization:
					   vector<double> &randScores);



int compareStringDoubleMapValsDesc(const void *pIt1, const void *pIt2);
int compareStringDoubleMapVals(const void *pIt1, const void *pIt2);

bool compareUnsignedInt2DoubleDesc(const pair<unsigned int, double> &v1,
				   const pair<unsigned int, double> &v2);

bool compareString2DoubleDesc(const pair<string, double> &v1, const pair<string, double> &v2);
bool compareString2Double(const pair<string, double> &v1, const pair<string, double> &v2);



double computeNormalizedPearson(vector<double> &exprProfile1,
				vector<double> &exprProfile2);
double computeNormalizedPearson(vector<double> &exprProfile1,
				vector<double> &exprProfile2,
				vector<double> &weigths,
				void *functSpecInput);

double computePearson(vector<double> &exprProfile1,
		      vector<double> &exprProfile2);
double computePearson(vector<double> &exprProfile1,
		      vector<double> &exprProfile2,
		      vector<double> &weigths,
		      void *functSpecInput);

double computeQuadraticPearson(vector<double> &exprProfile1,
			       vector<double> &exprProfile2);
double computeQuadraticPearson(vector<double> &exprProfile1,
			       vector<double> &exprProfile2,
			       vector<double> &weigths,
			       void *functSpecInput);

double computeInverseEucDist(vector<double> &exprProfile1,
			     vector<double> &exprProfile2);
double computeInverseEucDist(vector<double> &exprProfile1,
			     vector<double> &exprProfile2,
			     vector<double> &weigths,
			     void *functSpecInput);

double computeEucDist(vector<double> &exprProfile1,
		      vector<double> &exprProfile2);
double computeEucDist(vector<double> &exprProfile1,
		      vector<double> &exprProfile2,
		      vector<double> &weigths,
		      void *functSpecInput);

double computeCosine(vector<double> &exprProfile1,
		     vector<double> &exprProfile2);
double computeCosine(vector<double> &exprProfile1,
		     vector<double> &exprProfile2,
		     vector<double> &weigths,
		     void *functSpecInput);

//computeSpearman not needed -> pearson on rank transformed data!

double computeRelativeIntensityOverlap(vector<double> &exprProfile1,
				       vector<double> &exprProfile2);
double computeRelativeIntensityOverlap(vector<double> &exprProfile1,
				       vector<double> &exprProfile2,
				       vector<double> &weigths,
				       void *functSpecInput);

double computePearson3D(vector<double> &exprProfile1,
			vector<double> &exprProfile2,
			vector<double> &weights, // weights for planes!
			void *functSpecInput);

double computeAveragePairwiseScore(vector<string> &genes,
			   map<string,vector<double> > &exprProfileMap,
			   double (*computeCoExprTwoGenes)(vector<double> &,
							   vector<double> &,
							   vector<double> &,
							   void *),
			   vector<double> &weights,
			   void *functSpecInput);

int getStdOutFromCommand(const string command, string &out);

// template <class T>
// bool addToVectorUnique(vector<T> &vect, const T toAdd);

#endif /* #ifndef _CGPACE_SERVICEROUTINES_H_ */
