/*****************************************************************************
 * mbc-evalCandGenesByCoExpr.cpp - Source file for CGPACE - evaluation of    *
 * candidate genes, given a list of candidates, a list of known disease and  *
 * a list of Pearson correlation coefficients (or a similar measure for      *
 * co-expression; having higher values for positive co-expression).          *
 * Version 0.0.6                                                             *
 * ------------------------------------------------------------------------- *
 * Copyright 2007-2012 by Rosario Piro (mailto: rmpiro@gmail.com)            *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; _exclusively_ version 2 of the License    *
 *   NO later version.                                                       *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *   or see http://www.gnu.org/copyleft/gpl.html                             *
 * ------------------------------------------------------------------------- */

//---------------------
//--- include files ---
//---------------------
//#include <fstream>
#include <iostream>
//#include <map>
//#include <set>
//#include <vector>
#include <iomanip>
//#include <cmath>
#include <cstring>
#include <algorithm>
//#include <cstdlib>
//#include <climits>
#include "extstdlib.h"
#include "evalCandGenesByCoExpr-serviceroutines.h"
//#include <ctime>

#ifdef HAVE_GSL
  #include <gsl/gsl_rng.h>
#else
  #include <cstdlib>
#endif

//---------------------------
//--- preprocessor macros ---
//---------------------------
#define PRGVERSION "0.0.6"
#define GSL_RANDOMGENERATOR gsl_rng_ranlxs2

#define NO_EXPRVALUE DBL_MIN

//-----------------------------/
//--- function declarations ---/
//-----------------------------/

void printUsageInfo(const char *prgname);

//inline bool invDoubleComp(double i, double j) { return (i > j); }


//------------------------/
//--- global variables ---/
//------------------------/

//------------
//--- main ---
//------------

int main (int argc, const char **argv) {

  //cerr << "TODO: implement usage of DGP scores!" << endl << flush;

  int exitStatus = 0;

  #ifdef HAVE_GSL
    gsl_rng *rgen;
  #endif

  bool useDGPprob = false;     // default
  bool extendedOutput = false; // default
  bool coexprLists4CandGenes = false; // default

  vector<string> candGenes;    // list of candidate genes to be ranked
  vector<string> refGenes;     // list of reference genes

  set<string> genesWithExprData;

  map<string,vector<double> > exprProfileMap;

  map<string,map<string,double> > coexprListsRankMap;
                               // maps coexpr. lists to a gene-(rel.)rank map; we use this, in order
                               // not to re-compute always the same rankings when randomizing

  vector<string> headerInfo;
  map<string,vector<unsigned int> > planes;
  void *functSpecInfo = NULL;
  double minNormExpr = DBL_MIN; // used only for "relintoverlap"; default: no
                                // minimum/threshold.

  vector<double> weights; // default: empty, no weights used!

  double (*computeCoExprTwoGenes)(vector<double> &,
				  vector<double> &,
				  vector<double> &,
				  void *) = computePearson;
                               // points to the function that determins
                               // the co-expression of two genes.
                               // Default: Pearson correlation coefficient
  string coExprScoringFunction = "pearson";

  map<string,double> geneDGPmap; // maps genes to their DGP value
                                 // (Lopez-Bigas et al.)

  vector<string> rankedCandGenes;           // final ranked list of candidates.
  map<string,double> candGene2ScoreMap;     // to store overall scores.
  map<string,double> candGene2AvCoexScoreMap; // to store average CoexScores.

  map<string,double> gene2FDR;              // used if we randomize (option -n) ...
  bool randomizeRefGenes = false;           // default: randomize candidate genes

  map<string,vector<double> > candGene2relPositionsMap; // for keeping track
                                          // of the relative positions the
                                          // candidates had in the reference
                                          // gene co-expression lists or vice
                                          // versa; for extended output.

  string candGeneFileName = "";
  string refGeneFileName = "";
  string weightsFileName = "";
  //string pearsonFileName = "";
  string exprDataFileName = "";
  string dgpFileName = "";
  //string outputFileName = "";

  int numRand = 0; // default: no randomizations and FDR estimation
  vector<double> randScores;
  bool withPseudoCount = false;

  char *dateString;

  cout << setprecision(DEFAULT_PRECISION);

  // parse command line arguments:
  int index = 1;
  while (index < argc) {
    // help
    if (strcmp(argv[index],"--help")==0 || strcmp(argv[index],"-h")==0) {
      printUsageInfo(argv[0]);
      exit(0);
    }
    // candidate gene file
    else if (strcmp(argv[index],"-c")==0) {
      if (index + 1 < argc) {
	candGeneFileName = argv[++index];
      }
      else { // filename missing after -c
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // reference gene file
    else if (strcmp(argv[index],"-r")==0) {
      if (index + 1 < argc) {
	refGeneFileName = argv[++index];
      }
      else { // filename missing after -r
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // Expression data file
    else if (strcmp(argv[index],"-e")==0) {
      if (index + 1 < argc) {
	exprDataFileName = argv[++index];
      }
      else { // filename missing after -e
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // Weight file
    else if (strcmp(argv[index],"-w")==0) {
      if (index + 1 < argc) {
	weightsFileName = argv[++index];
      }
      else { // filename missing after -w
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // coexpression score for two genes:
    else if (strcmp(argv[index],"-s")==0) {
      if (index + 1 < argc) {
	++index;
	if (strcmp(argv[index],"pearson")==0) {
	  computeCoExprTwoGenes = computePearson;
	  coExprScoringFunction = "pearson";
	}
	else if (strcmp(argv[index],"quadraticpearson")==0) {
	  computeCoExprTwoGenes = computeQuadraticPearson;
	  coExprScoringFunction = "quadraticpearson";
	}
	else if (strcmp(argv[index],"euclidean")==0) {
	  computeCoExprTwoGenes = computeInverseEucDist;
	  coExprScoringFunction = "euclidean";
	}
	else if (strcmp(argv[index],"cosine")==0) {
	  computeCoExprTwoGenes = computeCosine;
	  coExprScoringFunction = "cosine";
	}
	else if (strcmp(argv[index],"spearman")==0) {
	  computeCoExprTwoGenes = computePearson;  // ! spearman = pearson on
	                                           //   rank transformed data!
	  coExprScoringFunction = "spearman";
	}
	else if (strcmp(argv[index],"relintoverlap")==0) {
	  computeCoExprTwoGenes = computeRelativeIntensityOverlap;
	  coExprScoringFunction = "relintoverlap";
	}
	//else if (strcmp(argv[index],"normpearson")) {
	//computeCoExprTwoGenes = computeNormalizedPearson;
	//coExprScoringFunction = "normpearson";
	//}
	else if (strcmp(argv[index],"pearson3d")==0) {
	  computeCoExprTwoGenes = computePearson3D;
	  coExprScoringFunction = "pearson3d";
	}
	else { // unknwon scoring function
	  printUsageInfo(argv[0]);
	  exit(1);
	}
      }
      else { // scoring function missing after -s
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // build coexpression lists for candidates instead of reference genes? 
    else if (strcmp(argv[index],"-C")==0) {
      coexprLists4CandGenes = true;
    }
    // use minimum normalized expression for "relintoverlap"
    else if (strcmp(argv[index],"-M")==0) {
      if (index + 1 < argc) {
	const char *val = argv[++index];
	if (isNumeric(val))
	  minNormExpr = atof(val);
	else {
	  printUsageInfo(argv[0]);
	  exit(1);
	}
      }
      else { // threshold missing after -M
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // use randomizations for FDR estimation
    else if (strcmp(argv[index],"-n")==0) {
      if (index + 1 < argc) {
	const char *val = argv[++index];
	if (isNumeric(val))
	  numRand = atoi(val);
	else {
	  printUsageInfo(argv[0]);
	  exit(1);
	}
      }
      else { // number missing after -n
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // use pseudocount?
    else if (strcmp(argv[index],"-P")==0) {
      withPseudoCount = true;
    }
    // randomize ref. genes instead of cand. genes?
    else if (strcmp(argv[index],"-R")==0) {
      randomizeRefGenes = true;
    }

    // output file
    //else if (strcmp(argv[index],"-o")==0) {
    //  if (index + 1 < argc) {
    //outputFileName = argv[++index];
    //}
    //else { // filename missing after -o
    //printUsageInfo(argv[0]);
    //exit(1);
    //}
    //}
    // use DGP probabilites
    else if (strcmp(argv[index],"-d")==0) {
      if (index + 1 < argc) {
	dgpFileName = argv[++index];
	useDGPprob = true;
      }
      else { // filename missing after -d
	printUsageInfo(argv[0]);
	exit(1);
      }
    }
    // extended output
    else if (strcmp(argv[index],"-E")==0) {
      extendedOutput = true;
    }

    index++;
  }

  // check for mandatory parameters:
  if (candGeneFileName == "" || refGeneFileName == ""
      || exprDataFileName == "" ) {
    printUsageInfo(argv[0]);
    exit(1);
  }


  // before starting determine date and time:
  time_t currTime = time(NULL);
  dateString = ctime(&currTime);


  // initialize random generator.
  #ifdef HAVE_GSL
    rgen = gsl_rng_alloc(GSL_RANDOMGENERATOR);
    gsl_rng_set(rgen, currTime);
  #else
    srand(currTime);
  #endif


  // read input file for candidate genes:
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: candidate genes ==========] " << endl;
  #endif

  if ( (exitStatus=parseGeneFile(candGeneFileName, candGenes)) != 0) {
    cerr << "Error: Couldn't read file " << candGeneFileName
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }

  
  // read input file for reference genes:
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: reference genes ==========] " << endl;
  #endif

  if ( (exitStatus=parseGeneFile(refGeneFileName, refGenes)) != 0) {
    cerr << "Error: Couldn't read file " << refGeneFileName
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }


  if (weightsFileName != "") {
    // read input file for weight vector:
    #ifdef DEBUG
    cout << "#DEBUG: [========== File: weight vector ==========] " << endl;
    #endif

    if ( (exitStatus=parseWeightFile(weightsFileName, weights)) != 0) {
      cerr << "Error: Couldn't read file " << weightsFileName
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }
  }


  // read expression data file (expression profiles):
  #ifdef DEBUG
  cout << "#DEBUG: [========== File: Expression data ==========] " << endl;
  #endif

  if ( (exitStatus=parseExpressionDataFile(exprDataFileName, genesWithExprData,
					   exprProfileMap, headerInfo)) != 0) {
    cerr << "Error: Couldn't read file " << exprDataFileName
	 << " ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }

  // set additional input information or perform additional pre-processing
  // for some of the correlation measures:

  // if needed, parse header info to extract planes:
  if (coExprScoringFunction == "pearson3d") {
    if ((exitStatus=convertHeaderInfo2Planes(headerInfo, planes)) != 0) {
      cerr << "Error: Couldn't convert header info to planes "
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }

    functSpecInfo = &planes; // need this to pass
  }
  else if (coExprScoringFunction == "relintoverlap") {
    // if we don't need planes, we may need a minimum normalized expr.
    // (for noise reduction with relintoverlap)
    functSpecInfo = &minNormExpr;
  }
  else if (coExprScoringFunction == "spearman") {
    // Spearman rank correlation = Pearson correlation on rank transformed data
    #ifdef DEBUG
    cout << "#DEBUG: [========== Rank transformation of expression data for Spearman ==========] " << endl;
    #endif

    
    map<string,vector<double> >::iterator ePMit = exprProfileMap.begin();
    unsigned int ePlength = ePMit->second.size();

    vector<pair<unsigned int, double> > idx2valueVect;
    idx2valueVect.resize(ePlength);


    for ( ; ePMit != exprProfileMap.end(); ePMit++) {
      // for each expression profile ...

      #ifdef DEBUG
      cout << "#DEBUG: expression profile for " << ePMit->first << endl;
      cout << "#DEBUG:    profile:";
      for (vector<double>::iterator vit = ePMit->second.begin();
	   vit != ePMit->second.end(); vit++)
	cout << "\t" << (*vit);
      cout << flush << endl;
      #endif
      
      // step 1: copy the expression profile with index indormation
      unsigned int index;
      for (index = 0; index < ePlength; index++) {
	idx2valueVect[index].first = index;                 // original index
	idx2valueVect[index].second = ePMit->second[index]; // value
      }

      // step 2: sort the profile according to decreasing expr. values
      sort(idx2valueVect.begin(), idx2valueVect.end(),
	   compareUnsignedInt2DoubleDesc);

      // step 3: replace the original profile by ranks
      unsigned int sumRank = 0;
      unsigned int numRanksWithValue = 0;
      double lastValue = NO_EXPRVALUE;
      vector<double> origIndicesWithValue;

      unsigned int thisRank = 0;
      vector<pair<unsigned int, double> >::iterator i2vIt;
      for (i2vIt = idx2valueVect.begin(); i2vIt != idx2valueVect.end();
	   i2vIt++) {

	thisRank++;
	unsigned int orig_index = i2vIt->first;
	double thisValue = i2vIt->second;

	if (thisValue == lastValue) {
	  sumRank += thisRank;  // same value, need to average over the ranks
	  numRanksWithValue++;  // later!
	  origIndicesWithValue.push_back(orig_index);

          //#ifdef DEBUG
	  //cout << "#DEBUG: SAME VALUE! thisRank=" << thisRank << "; orig_index="
	  //     << orig_index << "; thisValue=" << thisValue
	  //     << "; lastValue="<< lastValue << "; sumRank=" << sumRank
	  //     << "; numRanksWithValue=" << numRanksWithValue << flush << endl;
          //#endif
	}
	else {
	  // write the previous average (if we had more data points with the
	  // same value)
	  vector<double>::iterator oIit;
	  for (oIit = origIndicesWithValue.begin();
	       oIit != origIndicesWithValue.end(); oIit++) {
	    ePMit->second[*oIit] = (double)sumRank/(double)numRanksWithValue;
	  }
	  origIndicesWithValue.clear();

	  // this is a new value
	  sumRank = thisRank;
	  numRanksWithValue = 1;

	  //#ifdef DEBUG
	  //cout << "#DEBUG: NEW VALUE! thisRank=" << thisRank << "; orig_index="
	  //     << orig_index << "; thisValue=" << thisValue
	  //     << "; lastValue="<< lastValue << "; sumRank=" << sumRank
	  //     << "; numRanksWithValue=" << numRanksWithValue << flush << endl;
          //#endif

	  lastValue = thisValue;
	  origIndicesWithValue.push_back(orig_index);
	}


      } // for (index = 0; ...

      // write also the last!
      vector<double>::iterator oIit;
      for (oIit = origIndicesWithValue.begin();
	   oIit != origIndicesWithValue.end(); oIit++) {
	ePMit->second[*oIit] = (double)sumRank/(double)numRanksWithValue;
      }
      origIndicesWithValue.clear();


      #ifdef DEBUG
      cout << "#DEBUG:" << endl;
      cout << "#DEBUG:     ranked:";
      for (vector<double>::iterator vit = ePMit->second.begin();
	   vit != ePMit->second.end(); vit++)
	cout << "\t" << (*vit);
      cout << endl;
      cout << "#DEBUG:" << endl;
      #endif

    } // for ( ; ePMit ...



  } // else if (coExprScoringFunction == "spearman") ...


  // DO SOME STATISTICS:
  // - fraction of reference genes in expression data?
  // - fraction of candidate genes in expression data?
  // - average pairwise correlation score of reference genes?
  unsigned int numRefGenesPresent = 0;
  vector<string> presentRefGenes;

  unsigned int rG;
  for (rG = 0; rG < refGenes.size(); rG++) {
    if (genesWithExprData.find(refGenes[rG]) != genesWithExprData.end()) {
      numRefGenesPresent++;
      presentRefGenes.push_back(refGenes[rG]);
    }
  }

  double fracRefGenesWithExprData =
    (double)numRefGenesPresent / (double)refGenes.size();

  
  int numCandGenesPresent = 0;
  vector<string> presentCandGenes;
  unsigned int cG;
  for (cG = 0; cG < candGenes.size(); cG++) {
    if (genesWithExprData.find(candGenes[cG]) != genesWithExprData.end()) {
      numCandGenesPresent++;
      presentCandGenes.push_back(candGenes[cG]);
    }
  }

  double fracCandGenesWithExprData =
    (double)numCandGenesPresent / (double)candGenes.size();


  double avRefCorrScore = computeAveragePairwiseScore(presentRefGenes,
						      exprProfileMap,
						      computeCoExprTwoGenes,
						      weights, functSpecInfo);


  if (useDGPprob) {
    // read DGP data dile:
    #ifdef DEBUG
    cout << "#DEBUG: [========== File: DGP data ==========] " << endl;
    #endif

    if ( (exitStatus=parseDGPDataFile(dgpFileName, geneDGPmap)) != 0) {
      cerr << "Error: Couldn't read file " << dgpFileName
	   << " ... quitting! Errorcode: " << exitStatus << endl;
      exit(exitStatus);
    }
  }



  // rank the true candidates for the true reference gene set:
  #ifdef DEBUG
  cout << "#DEBUG: [========== Evaluating candidate genes for reference gene set ==========] " << endl;
  #endif

  vector<double> randScoresDummy; // remains empty
  if ( (exitStatus=evaluateCandidateList4ReferenceGeneSet(candGenes, refGenes,
							  exprProfileMap, weights, functSpecInfo,
							  useDGPprob, geneDGPmap,
							  coexprLists4CandGenes,
							  coexprListsRankMap,
							  computeCoExprTwoGenes,
							  false, extendedOutput,  // first is: isRandomization
							  // output:
							  rankedCandGenes,
							  candGene2ScoreMap,
							  candGene2AvCoexScoreMap,
							  // extended output:
							  candGene2relPositionsMap,
							  // rand. output
							  randScoresDummy)) != 0) {
    cerr << "Error: Couldn't evaluate candidate genes for the reference gene set ... quitting! Errorcode: " << exitStatus << endl;
    exit(exitStatus);
  }


  if (numRand > 0) {
    // we need to estimate FDRs; compute random scores:
    #ifdef DEBUG
    cout << "#DEBUG: [========== Computing random scores for randomized gene sets ==========] " << endl;
    #endif

    vector<string> randCandGenes = candGenes;
    vector<string> randRefGenes = refGenes;

    unsigned int randGeneSetSize = 0;

    if (randomizeRefGenes) { // option -R; randomize reference gene set; not candidate genes
      randGeneSetSize = numRefGenesPresent;
    }
    else { // default: randomize candidate gene set
      randGeneSetSize = numCandGenesPresent;
    }


    for (int thisRand = 0; thisRand < numRand; thisRand++) { // new random gene set

      #ifdef DEBUG
      cout << "#DEBUG: Random gene set number " << (thisRand+1) << endl;
      #endif

      // generate random gene set from genes with expression data:
      vector<string> randGenes;
      set<string> genesWithExprDataTmp = genesWithExprData;

      while (!genesWithExprDataTmp.empty() && randGenes.size()<randGeneSetSize) {
	// not empty, randomly pick a value and add it to the randomized gene set until the gene set has the correct size

        #ifdef HAVE_GSL
          int ranIndex = (int) ( gsl_rng_uniform(rgen) * genesWithExprDataTmp.size() );
        #else
          int ranIndex = (int) ( ((double)rand()/(RAND_MAX+1.0)) * genesWithExprDataTmp.size() );
        #endif

	// add the randomly chosen expression value to randomized gene set and remove it from the total number of genes (drawing without replacement!)
	set<string>::iterator sit = genesWithExprDataTmp.begin();
	for (int ii=0; ii<ranIndex; ii++) {
	  sit++;
	}
	randGenes.push_back(*sit);
	genesWithExprDataTmp.erase(*sit);
      }


      if (randomizeRefGenes) { // option -R; randomize reference gene set; not candidate genes
	randRefGenes = randGenes;
      }
      else { // default: randomize candidate gene set
	randCandGenes = randGenes;
      }


      // now compute scores for this ransomization set (appended to randScores)

      vector<string> rankedCandGenesDummy;   // these remain empty
      map<string,double> candGene2ScoreMapDummy;
      map<string,double> candGene2AvCoexScoreMapDummy;
      map<string,vector<double> > candGene2relPositionsMapDummy;

      if ( (exitStatus=evaluateCandidateList4ReferenceGeneSet(randCandGenes, randRefGenes,
							      exprProfileMap, weights, functSpecInfo,
							      useDGPprob, geneDGPmap,
							      coexprLists4CandGenes,
							      coexprListsRankMap,
							      computeCoExprTwoGenes,
							      true, extendedOutput,  // first is: isRandomization
							      // output:
							      rankedCandGenesDummy,
							      candGene2ScoreMapDummy,
							      candGene2AvCoexScoreMapDummy,
							      // extended output:
							      candGene2relPositionsMapDummy,
							      // rand. output
							      randScores)) != 0) {
	cerr << "Error: Couldn't compute scores for randomization ... quitting! Errorcode: " << exitStatus << endl;
	exit(exitStatus);
      }


    } // iteration over randomizations


    // sort the randomization results:
    #ifdef DEBUG
    cout << "#DEBUG: sorting true results numerically ... " << endl << flush;
    #endif

    sort(randScores.begin(), randScores.end());

    #ifdef DEBUG
    cout << "#DEBUG: top ten random scores are:" << endl;
    for (int ii=0; ii<10; ii++) {
      cout << "#DEBUG: " << randScores[ii] << endl;
    }
    #endif



    // we have all random scores; compute FDRs for true scores:
    #ifdef DEBUG
    cout << "#DEBUG: [========== False Discovery Rates (FDRs) ==========] " << endl;
    #endif


    for (unsigned int index = 0; index < rankedCandGenes.size(); index++) {

      string cGene = rankedCandGenes[index];
      double thisScore = candGene2ScoreMap.find(cGene)->second;

      //candGene2ScoreMap, rankedCandGenes
      #ifdef DEBUG
      cout << "#DEBUG: index=" << index << ": " << cGene << "; score: " << thisScore << "; " << flush;
      #endif

      unsigned int numBetterRandScores = 0;

      vector<double>::iterator rVecIt;
      for (rVecIt = randScores.begin(); rVecIt != randScores.end(); rVecIt++) {
        if (*rVecIt < thisScore) {
          numBetterRandScores++;
        }
      }

      #ifdef DEBUG
      cout << "number of better random scores: " << numBetterRandScores;
      if (withPseudoCount) {
        cout << "+1";
      }
      cout << " " << flush;
      #endif

      // with pseudocount?
      unsigned int pscount = 0;
      if (withPseudoCount) {
        pscount = 1;
      }
      double fdr = (double)(numBetterRandScores + pscount) / (double)(numRand*(index+1));


      #ifdef DEBUG
      cout << "=> fdr: " << fdr << endl << flush;
      #endif


      pair<string,double> newPair;
      newPair.first = cGene;
      newPair.second = fdr;

      if (gene2FDR.insert(newPair).second != true) {
        #ifdef DEBUG
        cout << "#DEBUG: Error: couldn't add FDR to list: "
             << cGene << endl;
        #endif
        exit(147);
      }

    } // for (unsigned int index = 0 ...


  } // if (numRand > 0


  // produce final output:
  #ifdef DEBUG
  cout << "#DEBUG: [========== Produce final ouptut ==========] " << endl;
  #endif

  cout << "# mbc-evalCandGenesByCoExpr" << endl;
  cout << "# Version " << PRGVERSION << endl;
  cout << "# Command: ";
  for (int jj = 0; jj < argc; jj++) {
    cout << argv[jj] << " ";
  }
  cout << endl;
  cout << "# Date: " << dateString;
  cout << "# Reference gene file: " << refGeneFileName << endl;
  cout << "# Number of reference genes: " << refGenes.size() << endl;

  cout << "# Number of reference genes having expression data: "
       << numRefGenesPresent << " (fraction="
       << fracRefGenesWithExprData << ")" << endl << flush;
  cout << "# Average pairwise correlation score of reference genes: ";
  if (avRefCorrScore == DBL_MIN) {
    cout << "-";
  }
  else {
    cout << avRefCorrScore;
  }
  cout << endl << flush;

  cout << "# Candidate gene file: " << candGeneFileName << endl;
  cout << "# Number of candidate genes: " << candGenes.size() << endl;

  cout << "# Number of candidate genes having necessary data for scoring (expression, DGP): "
       << numCandGenesPresent << " (fraction="
       << fracCandGenesWithExprData << ")" << endl << flush;

  cout << "# Expression data file: " << exprDataFileName << endl;
  cout << "# Scoring function for co-expression: " << coExprScoringFunction
       << endl;
  if (minNormExpr != DBL_MIN) {
    cout << "# Minimum normalized expression: " << minNormExpr
	 << endl;
  }
  if (weightsFileName != "") {
    cout << "# Weight vector for expression profiles: " << weightsFileName
	 << endl;
  }

  cout << "# Sorted co-expression lists built for: ";
  if (coexprLists4CandGenes)
    cout << "candidate genes";
  else
    cout << "reference genes";
  cout << endl;

  cout << "# Using DGP probabilities: ";
  if (useDGPprob)
    cout << "yes (file: " << dgpFileName << ")";
  else
    cout << "no";
  cout << endl;

  if (numRand > 0) {
    cout << "# Number of randomized gene sets for FDR estimation: " << numRand << endl;

    cout << "# Gene set to be randomized: ";
    if (randomizeRefGenes)
      cout << "reference genes";
    else
      cout << "candidate genes";
    cout << endl;

    #ifdef HAVE_GSL
      cout << "# Using random generator: " << gsl_rng_name(rgen)
           << " (GNU Scientific Library)" << endl;
    #else
      cout << "# Using random generator: system" << endl;
    #endif
  }


  // list of prioritized candidates:

  // header:
  cout << "rank\t"
       << "gene\t"
       << "score_mult\t"
       << "score_av\t"
       << "relRank_scored\t"
       << "relRank_all";
  if (useDGPprob)
    cout << "\tscore_dgp";
  if (numRand > 0)
    cout << "\tFDR";

  if (extendedOutput) {
    for (unsigned int jj = 0; jj < numRefGenesPresent; jj++) {
      cout << "\trelPos_" << presentRefGenes[jj];
    }
  }
  cout << endl;

  for (unsigned int candRank = 1; candRank <= rankedCandGenes.size();
       candRank++) {
    // - average Pearson for each candidate (either candidates or ref. genes in
    //   coexpression lists)
    // - relative rank of candidates regarding ALL candidates and regarding
    //   only scored candidates

    string thisCand = rankedCandGenes[candRank-1];
    double thisScore = (*(candGene2ScoreMap.find(thisCand))).second;

    double thisAvCoexScore =
                       (*(candGene2AvCoexScoreMap.find(thisCand))).second;
    double relRankAmongScoredCands = (double)candRank / 
                                                 (double)numCandGenesPresent;
    double relRankAmongAllCands = (double)candRank / 
                                                 (double)candGenes.size();

    cout << candRank << "\t"
	 << thisCand << "\t";

    if (thisScore == UNSCORED) {
      cout << "unscored\t"
	   << "unscored\t"
	   << "-" << "\t";
    }
    else {
      cout << thisScore << "\t"
	   << thisAvCoexScore << "\t"
	   << relRankAmongScoredCands << "\t";
    }

    cout << relRankAmongAllCands;

    if (useDGPprob) {
      map<string,double>::iterator gDGPit = geneDGPmap.find(thisCand);
      if (gDGPit == geneDGPmap.end())
	cout << "\tunscored";
      else
	cout << "\t" << (*gDGPit).second;
    }

    if (numRand > 0) { // we have randomized and write FDRs
      cout << "\t" << gene2FDR.find(thisCand)->second;
    }

    if (extendedOutput) {

      if (thisScore == UNSCORED) {
	for (unsigned int jj = 0; jj < numRefGenesPresent; jj++) {
	  cout << "\t-";
	}
      }
      else {
	map<string,vector<double> >::iterator rPit =
	  candGene2relPositionsMap.find(thisCand); // must exist!

	for (unsigned int jj = 0; jj < numRefGenesPresent; jj++) {
	  cout << "\t" << (*rPit).second[jj];
	}
      }
    }


    cout << endl;
  }


  exit(exitStatus);

} // main() ...



//----------------------------------------------------------------------

//------------------------/
//--- printUsageInfo() ---/
//------------------------/

void printUsageInfo(const char *prgname) {
  cerr << "Author: Rosario M. Piro (piro@to.infn.it)" << endl
       << "Usage: " << prgname << " [OPTIONS]" << endl << endl
       << "where options are:" << endl
       << " -h          Print this help message and quit." << endl
       << " -c <file>   File containing the list of candidate genes (one per line)." << endl
       << " -r <file>   File containing the list of reference genes (one per line)." << endl
       << " -e <file>   File containing the expression profiles (one profile per line; in" << endl
       << "             the form \"<gene>\\t<exprVal1>\\t<exprVal2>...\")."<< endl
       << " -w <file>   File containing a list of weights of the single experiments/data" << endl
       << "             points of the expression vectors; can be used for the scoring" << endl
       << "             functions: pearson, quadraticpearson, euclidean, cosine," << endl
       << "             spearman, relintoverlap, pearson3d (see below)." << endl
       << "             If not specified, no weighting will be used." << endl
       << " -s <func>   Scoring function to use for computing the co-expression of two" << endl
       << "             genes. Can be one of:" << endl
       << "             \"pearson\": Pearson correlation coefficient r (default)" << endl
    //<< "             \"normpearson\": normalized Pearson; for expression vectors with" << endl
    //<< "                            different numbers of expression values/experiments" << endl
       << "             \"quadraticpearson\": r*r (emphasizes higher values and considers" << endl
       << "                                 also anti-correlation)" << endl
       << "             \"spearman\": Spearman rank correlation" << endl
       << "             \"euclidean\": (inverse) Euclidean distance" << endl
       << "             \"cosine\": Cosine similarity" << endl    
       << "             \"relintoverlap\": Relative intensity overlap" << endl
       << "             \"pearson3d\": average of Pearson correlation coefficients" << endl
       << "                          computed for every plane of the 3D space. This" << endl
       << "                          requires an expression profile with positional" << endl
       << "                          information in the header; for each column: \"X,Y,Z\"." << endl
       << " -M <double> Minimum expression value. All" << endl
       << "             expression values below this threshold will be set to 0. This" << endl
       << "             is currently implemented only for the scoring function" << endl
       << "             \"relintoverlap\" and will be ignored otherwise." << endl
       << " -C          Build co-expression lists for the candidate genes instead of" << endl
       << "             the reference genes (check ranks of reference genes within the" << endl
       << "             co-expression lists of candidate genes instead of checking the" << endl
       << "             ranks of candidate genes within the reference genes' co-" << endl
       << "             expression lists)." << endl
       << " -d <file>   Integrate DGP (Lopez-Bigas et al.) disease gene probabilities" << endl
       << "             taken from the file. If not specified, DGP probilities are not" << endl
       << "             considered." << endl
    //       << " -o <file>   Output file for ranks of candidate genes" << endl
       << " -n <int>    Number of randomized candidate gene sets to evaluate for false discovery" << endl
       << "             rate (FDR) estimation. If not specified, no FDRs will be determined." << endl
       << " -R          Randomize the reference gene set instead of the candidate genes (used with -n)." << endl
       << " -P          Add a pseudo count of 1 when computing false discovery rates (so that they never become 0)." << endl
       << " -E          Extended output (relative positions in co-expression lists)." << endl << endl
       << "The options -c, -r, and -e are mandatory." << endl;
} // printUsageInfo() ...

